"""Contains core classes and functions for the Cleo package."""
from __future__ import annotations

import cleo.coords

# auto-import submodules
import cleo.ephys
import cleo.imaging
import cleo.ioproc
import cleo.opto
import cleo.recorders
import cleo.registry
import cleo.stimulators
import cleo.utilities
import cleo.viz
from cleo.base import (
    CLSimulator,
    InterfaceDevice,
    IOProcessor,
    Recorder,
    Stimulator,
    SynapseDevice,
)



"""Contains multi-unit and sorted spiking signals"""
from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Any, Tuple

import neo
import numpy as np
import quantities as pq
from attrs import define, field, fields
from bidict import bidict
from brian2 import NeuronGroup, Quantity, SpikeMonitor, meter, mm, ms
from nptyping import NDArray

from cleo.base import NeoExportable
from cleo.ephys.probes import Signal


@define(eq=False)
class Spiking(Signal, NeoExportable):
    """Base class for probabilistically detecting spikes"""

    r_perfect_detection: Quantity
    """Radius (with Brian unit) within which all spikes
    are detected"""
    r_half_detection: Quantity
    """Radius (with Brian unit) within which half of all spikes
    are detected"""
    cutoff_probability: float = 0.01
    """Spike detection probability below which neurons will not be
    considered. For computational efficiency."""
    t_ms: NDArray[(Any,), float] = field(
        init=False, factory=lambda: np.array([], dtype=float), repr=False
    )
    """Spike times in ms, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    i: NDArray[(Any,), np.uint] = field(
        init=False, factory=lambda: np.array([], dtype=np.uint), repr=False
    )
    """Channel (for multi-unit) or neuron (for sorted) indices
    of spikes, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    t_samp_ms: NDArray[(Any,), float] = field(
        init=False, factory=lambda: np.array([], dtype=float), repr=False
    )
    """Sample times in ms when each spike was recorded, stored if
    :attr:`~cleo.InterfaceDevice.save_history` on :attr:`~Signal.probe`"""
    i_probe_by_i_ng: bidict = field(init=False, factory=bidict, repr=False)
    """(neuron_group, i_ng) keys,  i_probe values. bidict for converting between
    neuron group indices and the indices the probe uses"""
    _monitors: list[SpikeMonitor] = field(init=False, factory=list, repr=False)
    _mon_spikes_already_seen: list[int] = field(init=False, factory=list, repr=False)
    _dtct_prob_array: NDArray[(Any, Any), float] = field(
        init=False, default=None, repr=False
    )

    def _init_saved_vars(self):
        if self.probe.save_history:
            self.t_ms = fields(type(self)).t_ms.default.factory()
            self.i = fields(type(self)).i.default.factory()
            self.t_samp_ms = fields(type(self)).t_samp_ms.default.factory()

    def _update_saved_vars(self, t_ms, i, t_samp_ms):
        if self.probe.save_history:
            self.i = np.concatenate([self.i, i])
            self.t_ms = np.concatenate([self.t_ms, t_ms])
            self.t_samp_ms = np.concatenate([self.t_samp_ms, [t_samp_ms]])

    def connect_to_neuron_group(
        self, neuron_group: NeuronGroup, **kwparams
    ) -> np.ndarray:
        """Configure signal to record from specified neuron group

        Parameters
        ----------
        neuron_group : NeuronGroup
            Neuron group to record from

        Returns
        -------
        np.ndarray
            num_neurons_to_consider x num_channels array of spike
            detection probabilities, for use in subclasses
        """
        super(Spiking, self).connect_to_neuron_group(neuron_group, **kwparams)
        # could support separate detection probabilities per group using kwparams
        # n_neurons X n_channels X 3
        dist2 = np.zeros((len(neuron_group), self.probe.n))
        for dim in ["x", "y", "z"]:
            dim_ng, dim_probe = np.meshgrid(
                getattr(neuron_group, dim),
                getattr(self.probe, f"{dim}s"),
                indexing="ij",
            )
            dist2 += (dim_ng - dim_probe) ** 2
        distances = np.sqrt(dist2) * meter  # since units stripped
        # probs is n_neurons by n_channels
        probs = self._detection_prob_for_distance(distances)
        # cut off to get indices of neurons to monitor
        # [0] since nonzero returns tuple of array per axis
        i_ng_to_keep = np.nonzero(np.all(probs > self.cutoff_probability, axis=1))[0]

        if len(i_ng_to_keep) > 0:
            # create monitor
            mon = SpikeMonitor(neuron_group)
            self._monitors.append(mon)
            self.brian_objects.add(mon)
            self._mon_spikes_already_seen.append(0)

            # update bidict from electrode group index to (neuron group, index)
            i_probe_start = len(self.i_probe_by_i_ng)
            new_i_probe = range(i_probe_start, i_probe_start + len(i_ng_to_keep))
            self.i_probe_by_i_ng.update(
                {
                    (neuron_group, i_ng): i_probe
                    for i_probe, i_ng in zip(new_i_probe, i_ng_to_keep)
                }
            )

        # return neuron-channel detection probabilities array for use in subclass
        return probs[i_ng_to_keep]

    @abstractmethod
    def get_state(
        self,
    ) -> tuple[NDArray[np.uint], NDArray[float], NDArray[np.uint]]:
        """Return spikes since method was last called (i, t_ms, y)

        Returns
        -------
        tuple[NDArray[np.uint], NDArray[float], NDArray[np.uint]]
            (i, t_ms, y) where i is channel (for multi-unit) or neuron (for sorted) spike
            indices, t_ms is spike times, and y is a spike count vector suitable for control-
            theoretic uses---i.e., a 0 for every channel/neuron that hasn't spiked and a 1
            for a single spike.
        """
        pass

    def _detection_prob_for_distance(self, r: Quantity) -> float:
        # p(d) = h/(r-c)
        a = self.r_perfect_detection
        b = self.r_half_detection
        # solving for p(a) = 1 and p(b) = .5 yields:
        c = 2 * a - b
        h = b - a
        with np.errstate(divide="ignore"):
            decaying_p = h / (r - c)
        decaying_p[decaying_p == np.inf] = 1  # to fix NaNs caused by /0
        p = 1 * (r <= a) + decaying_p * (r > a)
        return p

    def _i_ng_to_i_probe(self, i_ng, monitor):
        ng = monitor.source
        # assign value of -1 to neurons we aren't recording to filter out
        i_probe_unfilt = np.array([self.i_probe_by_i_ng.get((ng, k), -1) for k in i_ng])
        return i_probe_unfilt[i_probe_unfilt != -1].astype(np.uint)

    def _get_new_spikes(self) -> Tuple[npt.NDArray, npt.NDarray]:
        i_probe = np.array([], dtype=np.uint)
        t_ms = np.array([], dtype=float)
        for j in range(len(self._monitors)):
            mon = self._monitors[j]
            spikes_already_seen = self._mon_spikes_already_seen[j]
            i_ng = mon.i[spikes_already_seen:]  # can contain spikes we don't care about
            i_probe = np.concatenate((i_probe, self._i_ng_to_i_probe(i_ng, mon)))
            # get all time in terms of ms
            t_ms = np.concatenate((t_ms, mon.t[spikes_already_seen:] / ms))
            self._mon_spikes_already_seen[j] = mon.num_spikes

        return (i_probe, t_ms)

    def reset(self, **kwargs) -> None:
        # crucial that this be called after network restore
        # since that would reset monitors
        for j, mon in enumerate(self._monitors):
            self._mon_spikes_already_seen[j] = mon.num_spikes
        self._init_saved_vars()

    def to_neo(self) -> neo.Group:
        group = neo.Group(allowed_types=[neo.SpikeTrain])
        for i in set(self.i):
            st = neo.SpikeTrain(
                times=self.t_ms[self.i == i] * pq.ms,
                t_stop=self.probe.sim.network.t / ms * pq.ms,
            )
            st.annotate(i=int(i))
            group.add(st)

        group.annotate(export_datetime=datetime.now())
        group.name = f"{self.probe.name}.{self.name}"
        group.description = f"Exported from Cleo {self.__class__.__name__} object"
        return group


@define(eq=False)
class MultiUnitSpiking(Spiking):
    """Detects (unsorted) spikes per channel."""

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        """Configure signal to record from specified neuron group

        Parameters
        ----------
        neuron_group : NeuronGroup
            group to record from
        """
        neuron_channel_dtct_probs = super(
            MultiUnitSpiking, self
        ).connect_to_neuron_group(neuron_group, **kwparams)
        if self._dtct_prob_array is None:
            self._dtct_prob_array = neuron_channel_dtct_probs
        else:
            self._dtct_prob_array = np.concatenate(
                (self._dtct_prob_array, neuron_channel_dtct_probs), axis=0
            )

    def get_state(
        self,
    ) -> tuple[NDArray[np.uint], NDArray[float], NDArray[np.uint]]:
        # inherit docstring
        i_probe, t_ms = self._get_new_spikes()
        t_samp_ms = self.probe.sim.network.t / ms
        i_c, t_ms, y = self._noisily_detect_spikes_per_channel(i_probe, t_ms)
        self._update_saved_vars(t_ms, i_c, t_samp_ms)
        return (i_c, t_ms, y)

    def _noisily_detect_spikes_per_channel(
        self, i_probe, t
    ) -> Tuple[NDArray, NDArray, NDArray]:
        probs_for_spikes = self._dtct_prob_array[i_probe]
        detected_spikes = np.random.random(probs_for_spikes.shape) < probs_for_spikes
        y = np.sum(detected_spikes, axis=0)
        # ⬇ nonzero gives row, column indices of each nonzero element
        i_spike_detected, i_c_detected = detected_spikes.nonzero()
        t_detected = t[i_spike_detected]
        return i_c_detected, t_detected, y

    def to_neo(self) -> neo.Group:
        group = super(MultiUnitSpiking, self).to_neo()
        for st in group.spiketrains:
            i = int(st.annotations["i"])
            st.annotate(
                i_channel=i,
                x_contact=self.probe.coords[i, 0] / mm * pq.mm,
                y_contact=self.probe.coords[i, 1] / mm * pq.mm,
                z_contact=self.probe.coords[i, 2] / mm * pq.mm,
            )
        return group


@define(eq=False)
class SortedSpiking(Spiking):
    """Detect spikes identified by neuron indices.

    The indices used by the probe do not correspond to those
    coming from neuron groups, since the probe must consider
    multiple potential groups and within a group ignores those
    neurons that are too far away to be easily detected."""

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        """Configure sorted spiking signal to record from given neuron group

        Parameters
        ----------
        neuron_group : NeuronGroup
            group to record from
        """
        neuron_channel_dtct_probs = super(SortedSpiking, self).connect_to_neuron_group(
            neuron_group, **kwparams
        )
        neuron_multi_channel_probs = self._combine_channel_probs(
            neuron_channel_dtct_probs
        )
        if self._dtct_prob_array is None:
            self._dtct_prob_array = neuron_multi_channel_probs
        else:
            self._dtct_prob_array = np.concatenate(
                (self._dtct_prob_array, neuron_multi_channel_probs)
            )

    def _combine_channel_probs(self, neuron_channel_dtct_probs: NDArray) -> NDArray:
        # combine probabilities for each neuron to get just one representing
        # when a spike is detected on at least 1 channel
        # p(at least one detected) = 1 - p(none detected) = 1 - q1*q2*q3...
        return 1 - np.prod(1 - neuron_channel_dtct_probs, axis=1)

    def get_state(
        self,
    ) -> tuple[NDArray[np.uint], NDArray[float], NDArray[np.uint]]:
        # inherit docstring
        i_probe, t_ms = self._get_new_spikes()
        i_probe, t_ms = self._noisily_detect_spikes(i_probe, t_ms)
        y = np.zeros(len(self.i_probe_by_i_ng), dtype=bool)
        y[i_probe] = 1
        t_samp_ms = self.probe.sim.network.t / ms
        self._update_saved_vars(t_ms, i_probe, t_samp_ms)
        return (i_probe, t_ms, y)

    def _noisily_detect_spikes(self, i_probe, t) -> Tuple[NDArray, NDArray]:
        probs_for_spikes = self._dtct_prob_array[i_probe]
        detected_spikes = np.random.random(probs_for_spikes.shape) < probs_for_spikes
        i_spike_detected = detected_spikes.nonzero()
        i_probe_out = i_probe[i_spike_detected]
        t_out = t[i_spike_detected]
        return i_probe_out, t_out



from __future__ import annotations

import warnings
from typing import Callable, Tuple

from attrs import define, field
from brian2 import NeuronGroup, mm, np
from scipy.interpolate import CubicSpline

from cleo.coords import assign_xyz
from cleo.utilities import brian_safe_name, wavelength_to_rgb


def linear_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return np.interp(lambda_new_nm, lambdas_nm, epsilons)


def cubic_interpolator(lambdas_nm, epsilons, lambda_new_nm):
    return CubicSpline(lambdas_nm, epsilons)(lambda_new_nm)


# hacky MRO stuff...multiple inheritance only works because slots=False,
# and must be placed *before* SynapseDevice to work right
@define(eq=False, slots=False)
class LightDependent:
    """Mix-in class for opsin and light-dependent indicator.
    Light-dependent devices are connected to light sources (and vice-versa)
    on injection via the registry.

    We approximate dynamics under multiple wavelengths using a weighted sum
    of photon fluxes, where the ε factor indicates the activation
    relative to the peak-sensitivity wavelength for a equivalent power, which
    most papers report. When they report the action spectrum for equivalent
    photon flux instead (see Mager et al, 2018), use :func:`equal_photon_flux_spectrum`.
    This weighted sum is an approximation of a nonlinear
    peak-non-peak wavelength relation; see ``notebooks/multi_wavelength_model.ipynb``
    for details."""

    spectrum: list[tuple[float, float]] = field()
    """List of (wavelength, epsilon) tuples representing the action (opsin) or
    excitation (indicator) spectrum."""

    @spectrum.default
    def _default_spectrum(self):
        warnings.warn(
            f"No spectrum provided for light-dependent device {self.name}."
            " Assuming ε = 1 for all λ."
        )
        return [(-1e10, 1), (1e10, 1)]

    spectrum_interpolator: Callable = field(default=cubic_interpolator, repr=False)
    """Function of signature (lambdas_nm, epsilons, lambda_new_nm) that interpolates
    the action spectrum data and returns :math:`\\varepsilon \\in [0,1]` for the new
    wavelength."""

    @property
    def light_agg_ngs(self):
        return self.source_ngs

    def _get_source_for_synapse(
        self, target_ng: NeuronGroup, i_targets: list[int]
    ) -> Tuple[NeuronGroup, list[int]]:
        # create light aggregator neurons
        light_agg_ng = NeuronGroup(
            len(i_targets),
            model="""
            phi : 1/second/meter**2
            Irr : watt/meter**2
            """,
            name=f"light_agg_{brian_safe_name(self.name)}_{target_ng.name}",
        )
        assign_xyz(
            light_agg_ng,
            target_ng.x[i_targets] / mm,
            target_ng.y[i_targets] / mm,
            target_ng.z[i_targets] / mm,
            unit=mm,
        )
        return light_agg_ng, list(range(len(i_targets)))

    def epsilon(self, lambda_new) -> float:
        """Returns the :math:`\\varepsilon`` value for a given lambda (in nm)
        representing the relative sensitivity of the opsin to that wavelength."""
        action_spectrum = np.array(self.spectrum)
        lambdas = action_spectrum[:, 0]
        epsilons = action_spectrum[:, 1]
        if lambda_new < min(lambdas) or lambda_new > max(lambdas):
            warnings.warn(
                f"λ = {lambda_new} nm is outside the range of the action spectrum data"
                f" for {self.name}. Assuming ε = 0."
            )
            return 0
        eps_new = self.spectrum_interpolator(lambdas, epsilons, lambda_new)
        if eps_new < 0:
            warnings.warn(f"ε = {eps_new} < 0 for {self.name}. Setting ε = 0.")
            eps_new = 0
        return eps_new


def equal_photon_flux_spectrum(
    spectrum: list[tuple[float, float]],
) -> list[tuple[float, float]]:
    """Converts an equival photon flux spectrum to an equal power density spectrum."""
    spectrum = np.array(spectrum)
    lambdas = spectrum[:, 0]
    eps_phi = spectrum[:, 1]
    eps_Irr = eps_phi / lambdas
    eps_Irr /= np.max(eps_Irr)
    return list(zip(lambdas, eps_Irr))


def plot_spectra(*ldds: LightDependent) -> tuple[plt.Figure, plt.Axes]:
    """Plots the action/excitation spectra for multiple light-dependent devices."""
    import matplotlib.pyplot as plt

    fig, ax = plt.subplots()
    for ldd in ldds:
        spectrum = np.array(ldd.spectrum)
        lambdas = spectrum[:, 0]
        epsilons = spectrum[:, 1]
        lambdas_new = np.linspace(min(lambdas), max(lambdas), 100)
        epsilons_new = ldd.spectrum_interpolator(lambdas, epsilons, lambdas_new)
        c_points = [wavelength_to_rgb(l) for l in lambdas]
        c_line = wavelength_to_rgb(lambdas_new[np.argmax(epsilons_new)])
        ax.plot(lambdas_new, epsilons_new, c=c_line, label=ldd.name)
        ax.scatter(lambdas, epsilons, marker="o", s=50, color=c_points)
    title = (
        "Action/excitation spectra" if len(ldds) > 1 else f"Action/excitation spectrum"
    )
    ax.set(xlabel="λ (nm)", ylabel="ε", title=title)
    fig.legend()
    return fig, ax



"""Contains Light device and propagation models"""
from __future__ import annotations

import datetime
import warnings
from abc import ABC, abstractmethod
from typing import Any, Union

import matplotlib as mpl
import quantities as pq
from attrs import define, field
from brian2 import (
    NeuronGroup,
    Subgroup,
    np,
)
from brian2.units import (
    Quantity,
    mm,
    mm2,
    mwatt,
    nmeter,
    um,
)
from matplotlib import colors
from matplotlib.artist import Artist
from matplotlib.collections import PathCollection
from nptyping import NDArray

from cleo.base import CLSimulator
from cleo.coords import coords_from_xyz
from cleo.registry import registry_for_sim
from cleo.stimulators import Stimulator
from cleo.utilities import (
    analog_signal,
    normalize_coords,
    uniform_cylinder_rθz,
    wavelength_to_rgb,
    xyz_from_rθz,
)


@define
class LightModel(ABC):
    """Defines how light propagates given a source location and direction."""

    @abstractmethod
    def transmittance(
        self,
        source_coords: Quantity,
        source_direction: NDArray[(Any, 3), Any],
        target_coords: Quantity,
    ) -> NDArray[(Any, Any), float]:
        """Output must be between 0 and 1 with shape (n_sources, n_targets)."""
        pass

    @abstractmethod
    def viz_params(
        self,
        coords: Quantity,
        direction: NDArray[(Any, 3), Any],
        T_threshold: float,
        n_points_per_source: int = None,
        **kwargs,
    ) -> tuple[NDArray[(Any, Any, 3), Any], float, float]:
        """Returns info needed for visualization.
        Output is ((m, n_points_per_source, 3) viz_points array, markersize_um, intensity_scale).

        For best-looking results, implementations should scale `markersize_um` and `intensity_scale`.
        """
        pass

    def _get_rz_for_xyz(self, source_coords, source_direction, target_coords):
        """Assumes x, y, z already have units"""
        m = source_coords.reshape((-1, 3)).shape[0]
        if target_coords.ndim > 1:
            n = target_coords.shape[-2]
        elif target_coords.ndim == 1:
            n = 1
        # either shared or per-source targets
        assert target_coords.shape in [(m, n, 3), (1, n, 3), (n, 3), (3,)]

        # relative to light source(s)
        rel_coords = target_coords - source_coords.reshape((m, 1, 3))
        assert rel_coords.shape == (m, n, 3)
        # must use brian2's dot function for matrix multiply to preserve
        # units correctly.
        # zc = usf.dot(rel_coords, source_direction)  # mxn distance along cylinder axis
        #           m x n x 3    m x 1 x 3
        zc = np.sum(
            rel_coords * source_direction.reshape((-1, 1, 3)), axis=-1
        )  # mxn distance along cylinder axis
        assert zc.shape == (m, n)
        # just need length (norm) of radius vectors
        # not using np.linalg.norm because it strips units
        r = np.sqrt(
            np.sum(
                (
                    rel_coords
                    #    m x n                 m x 3
                    # --> m x n x 1             m x 1 x 3
                    - zc.reshape((m, n, 1)) * source_direction.reshape((-1, 1, 3))
                )
                ** 2,
                axis=-1,
            )
        )
        assert r.shape == (m, n)
        return r, zc


@define
class OpticFiber(LightModel):
    """Optic fiber light model from Foutz et al., 2012.

    Defaults are from paper for 473 nm wavelength."""

    R0: Quantity = 0.1 * mm
    """optical fiber radius"""
    NAfib: float = 0.37
    """optical fiber numerical aperture"""
    K: Quantity = 0.125 / mm
    """absorbance coefficient (wavelength/tissue dependent)"""
    S: Quantity = 7.37 / mm
    """scattering coefficient (wavelength/tissue dependent)"""
    ntis: float = 1.36
    """tissue index of refraction (wavelength/tissue dependent)"""

    def transmittance(
        self,
        source_coords: Quantity,
        source_dir_uvec: NDArray[(Any, 3), Any],
        target_coords: Quantity,
    ) -> NDArray[(Any, Any), float]:
        assert np.allclose(np.linalg.norm(source_dir_uvec, axis=-1), 1)
        r, z = self._get_rz_for_xyz(source_coords, source_dir_uvec, target_coords)
        return self._Foutz12_transmittance(r, z)

    def _Foutz12_transmittance(self, r, z, scatter=True, spread=True, gaussian=True):
        """Foutz et al. 2012 transmittance model: Gaussian cone with Kubelka-Munk propagation.

        r and z should be (n_source, n_target) arrays with units"""

        if spread:
            # divergence half-angle of cone
            theta_div = np.arcsin(self.NAfib / self.ntis)
            Rz = self.R0 + z * np.tan(
                theta_div
            )  # radius as light spreads ("apparent radius" from original code)
            C = (self.R0 / Rz) ** 2
        else:
            Rz = self.R0  # "apparent radius"
            C = 1

        if gaussian:
            G = 1 / np.sqrt(2 * np.pi) * np.exp(-2 * (r / Rz) ** 2)
        else:
            G = 1

        if scatter:
            S = self.S
            a = 1 + self.K / S
            b = np.sqrt(a**2 - 1)
            dist = np.sqrt(r**2 + z**2)
            M = b / (a * np.sinh(b * S * dist) + b * np.cosh(b * S * dist))
        else:
            M = 1

        T = G * C * M
        T[z < 0] = 0
        return T

    def viz_params(
        self,
        coords: Quantity,
        direction: NDArray[(Any, 3), Any],
        T_threshold: float,
        n_points_per_source: int = 4000,
        **kwargs,
    ) -> Quantity:
        r_thresh, zc_thresh = self._find_rz_thresholds(T_threshold)
        r, theta, zc = uniform_cylinder_rθz(n_points_per_source, r_thresh, zc_thresh)

        end = coords + zc_thresh * direction
        x, y, z = xyz_from_rθz(r, theta, zc, coords, end)
        density_factor = 3
        cyl_vol = np.pi * r_thresh**2 * zc_thresh
        markersize_um = (cyl_vol / n_points_per_source * density_factor) ** (1 / 3) / um
        intensity_scale = 1.5 * (4e3 / n_points_per_source) ** (1 / 3)
        return coords_from_xyz(x, y, z), markersize_um, intensity_scale

    def _find_rz_thresholds(self, thresh):
        """find r and z thresholds for visualization purposes"""
        res_mm = 0.01
        zc = np.arange(20, 0, -res_mm) * mm  # ascending T
        T = self._Foutz12_transmittance(0 * mm, zc)
        try:
            zc_thresh = zc[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            zc_thresh = np.max(zc)
        # look at half the z threshold for the r threshold
        r = np.arange(20, 0, -res_mm) * mm
        T = self._Foutz12_transmittance(r, zc_thresh / 2)
        try:
            r_thresh = r[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            r_thresh = np.max(r)
        # multiply by 1.2 just in case
        return r_thresh * 1.2, zc_thresh


def fiber473nm(
    R0=0.1 * mm,  # optical fiber radius
    NAfib=0.37,  # optical fiber numerical aperture
    K=0.125 / mm,  # absorbance coefficient
    S=7.37 / mm,  # scattering coefficient
    ntis=1.36,  # tissue index of refraction
) -> OpticFiber:
    """Returns an :class:`OpticFiber` model with parameters for 473 nm light.

    Parameters from Foutz et al., 2012."""
    return OpticFiber(
        R0=R0,
        NAfib=NAfib,
        K=K,
        S=S,
        ntis=ntis,
    )


@define
class Koehler(LightModel):
    """Even illumination over a circular area, with no scattering."""

    radius: Quantity
    """The radius of the Köhler beam"""
    zmax: Quantity = 500 * um
    """The maximum extent of the Köhler beam, 500 μm by default
    (i.e., no thicker than necessary to go through a slice or culture)."""

    def transmittance(self, source_coords, source_dir_uvec, target_coords):
        r, z = self._get_rz_for_xyz(source_coords, source_dir_uvec, target_coords)
        T = np.ones_like(r)
        T[r > self.radius] = 0
        T[z > self.zmax] = 0
        T[z < 0] = 0
        return T

    def viz_params(
        self, coords, direction, T_threshold, n_points_per_source=4000, **kwargs
    ):
        r, theta, zc = uniform_cylinder_rθz(n_points_per_source, self.radius, self.zmax)

        end = coords + self.zmax * direction
        x, y, z = xyz_from_rθz(r, theta, zc, coords, end)
        density_factor = 2
        cyl_vol = np.pi * self.radius**2 * self.zmax
        markersize_um = (cyl_vol / n_points_per_source * density_factor) ** (1 / 3) / um
        intensity_scale = (1 / n_points_per_source) ** (1 / 3)
        return coords_from_xyz(x, y, z), markersize_um, intensity_scale


@define(eq=False)
class Light(Stimulator):
    """Delivers light to the network for photostimulation and (when implemented) imaging.

    Requires neurons to have 3D spatial coordinates already assigned.

    Visualization kwargs
    --------------------
    n_points_per_source : int, optional
        The number of points per light source used to represent light intensity in
        space. Default varies by :attr:`light_model`. Alias ``n_points``.
    T_threshold : float, optional
        The transmittance below which no points are plotted. By default 1e-3.
    intensity : float, optional
        How bright the light appears, should be between 0 and 1. By default 0.5.
    rasterized : bool, optional
        Whether to render as rasterized in vector output, True by default.
        Useful since so many points makes later rendering and editing slow.
    """

    light_model: LightModel = field(kw_only=True)
    """Defines how light is emitted. See :class:`OpticFiber` for an example."""

    coords: Quantity = field(
        default=(0, 0, 0) * mm, converter=lambda x: np.reshape(x, (-1, 3)), repr=False
    )
    """(x, y, z) coords with Brian unit specifying where to place
    the base of the light source, by default (0, 0, 0)*mm.
    Can also be an nx3 array for multiple sources.
    """

    wavelength: Quantity = field(default=473 * nmeter, kw_only=True)
    """light wavelength with unit (usually nmeter)"""

    @coords.validator
    def _check_coords(self, attribute, value):
        if len(value.shape) != 2 or value.shape[1] != 3:
            raise ValueError(
                "coords must be an n by 3 array (with unit) with x, y, and z"
                "coordinates for n contact locations."
            )

    direction: NDArray[(Any, 3), Any] = field(
        default=(0, 0, 1), converter=normalize_coords
    )
    """(x, y, z) vector specifying direction in which light
    source is pointing, by default (0, 0, 1).
    
    Will be converted to unit magnitude."""

    max_Irr0_mW_per_mm2: float = field(default=None, kw_only=True)
    """The maximum irradiance the light source can emit.
    
    Usually determined by hardware in a real experiment."""

    max_Irr0_mW_per_mm2_viz: float = field(default=None, kw_only=True)
    """Maximum irradiance for visualization purposes. 
    
    i.e., the level at or above which the light appears maximally bright.
    Only relevant in video visualization.
    """

    default_value: NDArray[(Any,), float] = field(kw_only=True, repr=False)

    @default_value.default
    def _default_default(self):
        return np.zeros(self.n)

    @property
    def color(self):
        """Color of light"""
        return wavelength_to_rgb(self.wavelength / nmeter)

    def transmittance(self, target_coords) -> np.ndarray:
        """Returns :attr:`light_model` transmittance given light's coords and direction."""
        return self.light_model.transmittance(
            self.coords, self.direction, target_coords
        )

    def init_for_simulator(self, sim: CLSimulator) -> None:
        registry = registry_for_sim(sim)
        registry.init_register_light(self)
        self.reset()

    def connect_to_neuron_group(
        self, neuron_group: NeuronGroup, **kwparams: Any
    ) -> None:
        registry = registry_for_sim(self.sim)
        if self in registry.lights_for_ng.get(neuron_group, set()):
            raise ValueError(
                f"Light {self} already connected to neuron group {neuron_group}"
            )
        registry.register_light(self, neuron_group)

    @property
    def n(self):
        """Number of light sources"""
        assert len(self.coords.shape) == 2 or len(self.coords.shape) == 1
        return len(self.coords) if len(self.coords.shape) == 2 else 1

    @property
    def source(self) -> Subgroup:
        """Returns the "neuron(s)" representing the light source(s)."""
        registry = registry_for_sim(self.sim)
        return registry.source_for_light(self)

    def add_self_to_plot(self, ax, axis_scale_unit, **kwargs) -> list[PathCollection]:
        # show light with point field, assigning r and z coordinates
        # to all points
        # filter out points with <0.001 transmittance to make plotting faster

        # alias
        if "n_points" in kwargs and "n_points_per_source" not in kwargs:
            kwargs["n_points_per_source"] = kwargs.pop("n_points")
        T_threshold = kwargs.pop("T_threshold", 1e-3)

        viz_points, markersize_um, intensity_scale = self.light_model.viz_params(
            self.coords,
            self.direction,
            T_threshold,
            **kwargs,
        )
        assert viz_points.shape[0] == self.n
        assert viz_points.shape[2] == 3
        n_points_per_source = viz_points.shape[1]
        intensity = kwargs.get("intensity", 0.5 * intensity_scale)

        biggest_dim_pixels = max([ax.bbox.height, ax.bbox.width])
        dpi = 100  # default
        pt_per_in = 72
        biggest_dim_pt = biggest_dim_pixels / dpi * pt_per_in

        biggest_dim_um = (
            max(
                [
                    ax.get_xlim()[1] - ax.get_xlim()[0],
                    ax.get_ylim()[1] - ax.get_ylim()[0],
                    ax.get_zlim()[1] - ax.get_zlim()[0],
                ]
            )
            * axis_scale_unit
            / um
        )

        markersize_pt = markersize_um / biggest_dim_um * biggest_dim_pt
        markerarea = markersize_pt**2

        T = self.light_model.transmittance(self.coords, self.direction, viz_points)
        assert T.shape == (self.n, n_points_per_source)

        point_clouds = []
        for i in range(self.n):
            idx_to_plot = T[i] >= T_threshold
            point_cloud = ax.scatter(
                viz_points[i, idx_to_plot, 0] / axis_scale_unit,
                viz_points[i, idx_to_plot, 1] / axis_scale_unit,
                viz_points[i, idx_to_plot, 2] / axis_scale_unit,
                c=T[i, idx_to_plot],
                cmap=self._alpha_cmap_for_wavelength(intensity),
                vmin=0,
                vmax=1,
                marker="o",
                edgecolors="none",
                label=self.name,
                s=markerarea,
            )
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", message=".*Rasterization.*will be ignored.*"
                )
                # to make manageable in SVGs
                point_cloud.set_rasterized(kwargs.get("rasterized", True))
            point_clouds.append(point_cloud)

        handles = ax.get_legend().legendHandles
        c = wavelength_to_rgb(self.wavelength / nmeter)
        opto_patch = mpl.patches.Patch(color=c, label=self.name)
        handles.append(opto_patch)
        ax.legend(handles=handles)

        return point_clouds

    def update_artists(
        self, artists: list[Artist], value, *args, **kwargs
    ) -> list[Artist]:
        assert len(artists) == self.n
        if self.max_Irr0_mW_per_mm2_viz is not None:
            max_Irr0 = self.max_Irr0_mW_per_mm2_viz
        elif self.max_Irr0_mW_per_mm2 is not None:
            max_Irr0 = self.max_Irr0_mW_per_mm2
        else:
            raise Exception(
                f"Light'{self.name}' needs max_Irr0_mW_per_mm2_viz "
                "or max_Irr0_mW_per_mm2 "
                "set to visualize light intensity."
            )

        updated_artists = []
        for point_cloud, source_value in zip(artists, value):
            prev_value = getattr(point_cloud, "_prev_value", None)
            if source_value != prev_value:
                intensity = (
                    source_value / max_Irr0 if source_value <= max_Irr0 else max_Irr0
                )
                point_cloud.set_cmap(self._alpha_cmap_for_wavelength(intensity))
                updated_artists.append(point_cloud)
                point_cloud._prev_value = source_value

        return updated_artists

    def update(self, value: Union[float, np.ndarray]) -> None:
        """Set the light intensity, in mW/mm2 (without unit) for 1P
        excitation or laser power (mW) for 2P excitation (GaussianEllipsoid light model).

        Parameters
        ----------
        Irr0_mW_per_mm2 : float
            Desired light intensity for light source
        """
        if type(value) != np.ndarray:
            value = np.array(value).reshape((-1,))
        if value.shape not in [(), (1,), (self.n,)]:
            raise ValueError(
                f"Input to light must be a scalar or an array of"
                f" length {self.n}. Got {value.shape} instead."
            )
        if type(self.light_model) == "GaussianEllipsoid":
            # 10 microns, on upper end of what's used as spot size in Ronzitti et al., 2017
            # Irr0 = P / spot_area, as in Ronzitti et al., 2017
            cell_radius = 0.010  # mm
            cell_area = np.pi * cell_radius**2
            Irr0_mW_per_mm2 = value / cell_area
        else:
            Irr0_mW_per_mm2 = value
        if np.any(Irr0_mW_per_mm2 < 0):
            warnings.warn(f"{self.name}: negative light intensity Irr0 clipped to 0")
            Irr0_mW_per_mm2[Irr0_mW_per_mm2 < 0] = 0
        if self.max_Irr0_mW_per_mm2 is not None:
            Irr0_mW_per_mm2[
                Irr0_mW_per_mm2 > self.max_Irr0_mW_per_mm2
            ] = self.max_Irr0_mW_per_mm2
        super(Light, self).update(Irr0_mW_per_mm2)
        self.source.Irr0 = Irr0_mW_per_mm2 * mwatt / mm2

    def _alpha_cmap_for_wavelength(self, intensity):
        c = wavelength_to_rgb(self.wavelength / nmeter)
        c_dimmest = (*c, 0)
        alpha_max = 0.6
        alpha_brightest = alpha_max * intensity
        c_brightest = (*c, alpha_brightest)
        return colors.LinearSegmentedColormap.from_list(
            # breaks on newer matplotlib. leave 0 and 1 implicit
            # "incr_alpha", [(0, c_dimmest), (1, c_brightest)]
            "incr_alpha",
            [c_dimmest, c_brightest],
        )

    def to_neo(self):
        signal = analog_signal(self.t_ms, self.values, "mW/mm**2")
        signal.name = self.name
        signal.description = "Exported from Cleo Light device"
        signal.annotate(export_datetime=datetime.datetime.now())
        # broadcast in case of uniform direction
        _, direction = np.broadcast_arrays(self.coords, self.direction)
        signal.array_annotate(
            x=self.coords[..., 0] / mm * pq.mm,
            y=self.coords[..., 1] / mm * pq.mm,
            z=self.coords[..., 2] / mm * pq.mm,
            direction_x=direction[..., 0],
            direction_y=direction[..., 1],
            direction_z=direction[..., 2],
            i_channel=np.arange(self.n),
        )
        return signal



"""Contains Probe and Signal classes and electrode coordinate functions"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Tuple

import neo
from attrs import define, field
from brian2 import NeuronGroup, Quantity, Unit, mm, np, umeter
from matplotlib.artist import Artist
from mpl_toolkits.mplot3d.axes3d import Axes3D

from cleo.base import NeoExportable, Recorder
from cleo.coords import concat_coords
from cleo.utilities import get_orth_vectors_for_V


@define(eq=False)
class Signal(ABC):
    """Base class representing something an electrode can record"""

    name: str = field(kw_only=True)
    """Unique identifier used to organize probe output.
    Name of the class by default."""

    @name.default
    def _default_name(self) -> str:
        return self.__class__.__name__

    brian_objects: set = field(init=False, factory=set)
    """All Brian objects created by the signal.
    Must be kept up-to-date for automatic injection into the network"""
    probe: Probe = field(init=False, default=None)
    """The probe the signal is configured to record for."""

    def init_for_probe(self, probe: Probe) -> None:
        """Called when attached to a probe.

        Ensures signal can access probe and is only attached
        to one

        Parameters
        ----------
        probe : Probe
            Probe to attach to

        Raises
        ------
        ValueError
            When signal already attached to another probe
        """
        if self.probe is not None and self.probe is not probe:
            raise ValueError(
                f"Signal {self.name} has already been initialized "
                f"for Probe {self.probe.name} "
                f"and cannot be used with another."
            )
        self.probe = probe
        self._post_init_for_probe()

    def _post_init_for_probe(self):
        pass

    @abstractmethod
    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams):
        """Configure signal to record from specified neuron group

        Parameters
        ----------
        neuron_group : NeuronGroup
            group to record from
        """
        pass

    @abstractmethod
    def get_state(self) -> Any:
        """Get the signal's current value"""
        pass

    def reset(self, **kwargs) -> None:
        """Reset signal to a neutral state"""
        pass


@define(eq=False)
class Probe(Recorder, NeoExportable):
    """Picks up specified signals across an array of electrodes.

    Visualization kwargs
    --------------------
    marker : str, optional
        The marker used to represent each contact. "x" by default.
    size : float, optional
        The size of each contact marker. 40 by default.
    color : Any, optional
        The color of contact markers. "xkcd:dark gray" by default.
    """

    coords: Quantity = field(repr=False)
    """Coordinates of n electrodes. Must be an n x 3 array (with unit)
    where columns represent x, y, and z"""

    signals: list[Signal] = field(factory=list)
    """Signals recorded by the probe.
    Can be added to post-init with :meth:`add_signals`."""

    probe: Probe = field(init=False)

    def __attrs_post_init__(self):
        self.coords = self.coords.reshape((-1, 3))
        if len(self.coords.shape) != 2 or self.coords.shape[1] != 3:
            raise ValueError(
                "coords must be an n by 3 array (with unit) with x, y, and z"
                "coordinates for n contact locations."
            )
        for signal in self.signals:
            signal.init_for_probe(self)

    @property
    def n(self):
        """Number of electrode contacts in the probe"""
        return len(self.coords)

    def add_signals(self, *signals: Signal) -> None:
        """Add signals to the probe for recording

        Parameters
        ----------
        *signals : Signal
            signals to add
        """
        for signal in signals:
            signal.init_for_probe(self)
            self.signals.append(signal)

    def connect_to_neuron_group(
        self, neuron_group: NeuronGroup, **kwparams: Any
    ) -> None:
        """Configure probe to record from given neuron group

        Will call :meth:`Signal.connect_to_neuron_group` for each signal

        Parameters
        ----------
        neuron_group : NeuronGroup
            neuron group to connect to, i.e., record from
        **kwparams : Any
            Passed in to signals' connect functions, needed for some signals
        """
        for signal in self.signals:
            signal.connect_to_neuron_group(neuron_group, **kwparams)
            self.brian_objects.update(signal.brian_objects)

    def get_state(self) -> dict:
        """Get current state from probe, i.e., all signals

        Returns
        -------
        dict
            {'signal_name': value} dict with signal states
        """
        state_dict = {}
        for signal in self.signals:
            state_dict[signal.name] = signal.get_state()
        return state_dict

    def add_self_to_plot(
        self, ax: Axes3D, axis_scale_unit: Unit, **kwargs
    ) -> list[Artist]:
        # docstring inherited from InterfaceDevice
        marker = kwargs.get("marker", "x")
        size = kwargs.get("size", 40)
        color = kwargs.get("color", "xkcd:dark gray")

        markers = ax.scatter(
            self.xs / axis_scale_unit,
            self.ys / axis_scale_unit,
            self.zs / axis_scale_unit,
            marker=marker,
            s=size,
            color=color,
            label=self.name,
            depthshade=False,
        )
        handles = ax.get_legend().legendHandles
        handles.append(markers)
        ax.legend(handles=handles)
        return [markers]

    @property
    def xs(self) -> Quantity:
        """x coordinates of recording contacts

        Returns
        -------
        Quantity
            x coordinates represented as a Brian quantity, that is,
            including units. Should be like a 1D array.
        """
        return self.coords[:, 0]

    @property
    def ys(self) -> Quantity:
        """y coordinates of recording contacts

        Returns
        -------
        Quantity
            y coordinates represented as a Brian quantity, that is,
            including units. Should be like a 1D array.
        """
        return self.coords[:, 1]

    @property
    def zs(self) -> Quantity:
        """z coordinates of recording contacts

        Returns
        -------
        Quantity
            z coordinates represented as a Brian quantity, that is,
            including units. Should be like a 1D array.
        """
        return self.coords[:, 2]

    def reset(self, **kwargs):
        """Reset the probe to a neutral state

        Calls reset() on each signal
        """
        for signal in self.signals:
            signal.reset()

    def to_neo(self) -> neo.core.Group:
        group = neo.core.Group(
            name=self.name, description="Exported from Cleo Probe device"
        )
        for sig in self.signals:
            if not isinstance(sig, NeoExportable):
                continue
            group.add(sig.to_neo())
        return group


def linear_shank_coords(
    array_length: Quantity,
    channel_count: int,
    start_location: Quantity = (0, 0, 0) * mm,
    direction: Tuple[float, float, float] = (0, 0, 1),
) -> Quantity:
    """Generate coordinates in a linear pattern

    Parameters
    ----------
    array_length : Quantity
        Distance from the first to the last contact (with
        a Brian unit)
    channel_count : int
        Number of coordinates to generate, i.e. electrode contacts
    start_location : Quantity, optional
        x, y, z coordinate (with unit) for the start of the electrode
        array, by default (0, 0, 0)*mm
    direction : Tuple[float, float, float], optional
        x, y, z vector indicating the direction in which the array
        extends, by default (0, 0, 1), meaning pointing straight down

    Returns
    -------
    Quantity
        channel_count x 3 array of coordinates, where the 3 columns
        represent x, y, and z
    """
    dir_uvec = direction / np.linalg.norm(direction)
    end_location = start_location + array_length * dir_uvec
    return np.linspace(start_location, end_location, channel_count)


def tetrode_shank_coords(
    array_length: Quantity,
    tetrode_count: int,
    start_location: Quantity = (0, 0, 0) * mm,
    direction: Tuple[float, float, float] = (0, 0, 1),
    tetrode_width: Quantity = 25 * umeter,
) -> Quantity:
    """Generate coordinates for a linear array of tetrodes

    See https://www.neuronexus.com/products/electrode-arrays/up-to-15-mm-depth
    to visualize NeuroNexus-style arrays.

    Parameters
    ----------
    array_length : Quantity
        Distance from the center of the first tetrode to the
        last (with a Brian unit)
    tetrode_count : int
        Number of tetrodes desired
    start_location : Quantity, optional
        Center location of the first tetrode in the array,
        by default (0, 0, 0)*mm
    direction : Tuple[float, float, float], optional
        x, y, z vector determining the direction in which the linear
        array extends, by default (0, 0, 1), meaning straight down.
    tetrode_width : Quantity, optional
        Distance between contacts in a single tetrode. Not the diagonal
        distance, but the length of one side of the square. By default
        25*umeter, as in NeuroNexus probes.

    Returns
    -------
    Quantity
        (tetrode_count*4) x 3 array of coordinates, where 3 columns
        represent x, y, and z
    """
    dir_uvec = direction / np.linalg.norm(direction)
    end_location = start_location + array_length * dir_uvec
    center_locs = np.linspace(start_location, end_location, tetrode_count)
    # need to add coords around the center locations
    # tetrode_width is the length of one side of the square, so the diagonals
    # are measured in width/sqrt(2)
    #    x      -dir*width/sqrt(2)
    # x  .  x   +/- orth*width/sqrt(2)
    #    x      +dir*width/sqrt(2)
    orth_uvec, _ = get_orth_vectors_for_V(dir_uvec)
    return np.repeat(center_locs, 4, axis=0) + tetrode_width / np.sqrt(2) * np.tile(
        np.vstack([-dir_uvec, -orth_uvec, orth_uvec, dir_uvec]), (tetrode_count, 1)
    )


def poly2_shank_coords(
    array_length: Quantity,
    channel_count: int,
    intercol_space: Quantity,
    start_location: Quantity = (0, 0, 0) * mm,
    direction: Tuple[float, float, float] = (0, 0, 1),
) -> Quantity:
    """Generate NeuroNexus-style Poly2 array coordinates

    Poly2 refers to 2 parallel columns with staggered contacts.
    See https://www.neuronexus.com/products/electrode-arrays/up-to-15-mm-depth
    for more detail.

    Parameters
    ----------
    array_length : Quantity
        Length from the beginning to the end of the two-column
        array, as measured in the center
    channel_count : int
        Total (not per-column) number of coordinates (recording contacts) desired
    intercol_space : Quantity
        Distance between columns (with Brian unit)
    start_location : Quantity, optional
        Where to place the beginning of the array, by default (0, 0, 0)*mm
    direction : Tuple[float, float, float], optional
        x, y, z vector indicating the direction in which the two columns
        extend; by default (0, 0, 1), meaning straight down.

    Returns
    -------
    Quantity
        channel_count x 3 array of coordinates, where the 3 columns
        represent x, y, and z
    """
    dir_uvec = direction / np.linalg.norm(direction)
    end_location = start_location + array_length * dir_uvec
    out = np.linspace(start_location, end_location, channel_count)
    orth_uvec, _ = get_orth_vectors_for_V(dir_uvec)
    # place contacts on alternating sides of the central axis
    even_channels = np.arange(channel_count) % 2 == 0
    out[even_channels] += intercol_space / 2 * orth_uvec
    out[~even_channels] -= intercol_space / 2 * orth_uvec
    return out


def poly3_shank_coords(
    array_length: Quantity,
    channel_count: int,
    intercol_space: Quantity,
    start_location: Quantity = (0, 0, 0) * mm,
    direction: Tuple[float, float, float] = (0, 0, 1),
) -> Quantity:
    """Generate NeuroNexus Poly3-style array coordinates

    Poly3 refers to three parallel columns of electrodes.
    The middle column will be longest if the channel count
    isn't divisible by three and the side columns will be
    centered vertically with respect to the middle.

    Parameters
    ----------
    array_length : Quantity
        Length from beginning to end of the array as measured
        along the center column
    channel_count : int
        Total (not per-column) number of coordinates to generate
        (i.e., electrode contacts)
    intercol_space : Quantity
        Spacing between columns, with Brian unit
    start_location : Quantity, optional
        Location of beginning of the array, that is, the first contact
        in the center column, by default (0, 0, 0)*mm
    direction : Tuple[float, float, float], optional
        x, y, z vector indicating the direction along which the
        array extends, by default (0, 0, 1), meaning straight down

    Returns
    -------
    Quantity
        channel_count x 3 array of coordinates, where the 3 columns
        represent x, y, and z
    """
    # makes middle column longer if not even. Nothing fancier.
    # length measures middle column
    dir_uvec = direction / np.linalg.norm(direction)
    end_location = start_location + array_length * dir_uvec
    center_loc = start_location + array_length * dir_uvec / 2
    n_middle = channel_count // 3 + channel_count % 3
    n_side = int((channel_count - n_middle) / 2)

    middle = np.linspace(start_location, end_location, n_middle)

    spacing = array_length / n_middle
    side_length = n_side * spacing
    orth_uvec, _ = get_orth_vectors_for_V(dir_uvec)
    side = np.linspace(
        center_loc - dir_uvec * side_length / 2,
        center_loc + dir_uvec * side_length / 2,
        n_side,
    )
    side1 = side + orth_uvec * intercol_space
    side2 = side - orth_uvec * intercol_space
    out = concat_coords(middle, side1, side2)
    return out[out[:, 2].argsort()]  # sort to return superficial -> deep


def tile_coords(coords: Quantity, num_tiles: int, tile_vector: Quantity) -> Quantity:
    """Tile (repeat) coordinates to produce multi-shank/matrix arrays

    Parameters
    ----------
    coords : Quantity
        The n x 3 coordinates array to tile
    num_tiles : int
        Number of times to tile (repeat) the coordinates. For example,
        if you are tiling linear shank coordinates to produce multi-shank
        coordinates, this would be the desired number of shanks
    tile_vector : Quantity
        x, y, z array with Brian unit determining both the length
        and direction of the tiling

    Returns
    -------
    Quantity
        (n * num_tiles) x 3 array of coordinates, where the 3 columns
        represent x, y, and z
    """
    num_coords = coords.shape[0]
    # num_tiles X 3
    offsets = np.linspace((0, 0, 0) * mm, tile_vector, num_tiles)
    # num_coords X num_tiles X 3
    out = np.tile(coords[:, np.newaxis, :], (1, num_tiles, 1)) + offsets
    return out.reshape((num_coords * num_tiles, 3), order="F")



"""Classes and functions for constructing and configuring an :class:`~cleo.IOProcessor`."""
from cleo.ioproc.base import (
    LatencyIOProcessor,
    ProcessingBlock,
    RecordOnlyProcessor,
)
from cleo.ioproc.controllers import PIController
from cleo.ioproc.observers import FiringRateEstimator
from cleo.ioproc.delays import Delay, ConstantDelay, GaussianDelay



from __future__ import annotations

from attrs import define, field, fields_dict, asdict
from brian2 import Synapses, np, Quantity, NeuronGroup, second, umolar, nmolar

from cleo.base import SynapseDevice
from cleo.light import LightDependent


@define(eq=False)
class Sensor(SynapseDevice):
    """Base class for sensors"""

    sigma_noise: float = field(kw_only=True)
    """standard deviation of Gaussian noise in ΔF/F measurement"""
    dFF_1AP: float = field(kw_only=True)
    """ΔF/F for 1 AP, only used for scope SNR cutoff"""
    location: str = field(kw_only=True)
    """where sensor is expressed: cytoplasm or membrane"""

    @location.validator
    def _check_location(self, attribute, value):
        if value not in ("cytoplasm", "membrane"):
            raise ValueError(
                f"indicator location must be 'cytoplasm' or 'membrane', not {value}"
            )

    @property
    def snr(self) -> float:
        """Signal-to-noise ratio for 1 AP"""
        return self.dFF_1AP / self.sigma_noise

    def get_state(self) -> dict[NeuronGroup, np.ndarray]:
        """Returns a {neuron_group: fluorescence} dict of dFF values."""
        pass

    @property
    def exc_spectrum(self) -> list[tuple[float, float]]:
        """Excitation spectrum, alias for :attr:`spectrum`"""
        return self.spectrum


@define(eq=False)
class CalciumModel:
    """Base class for how GECI computes calcium concentration

    Must provide variable Ca (molar) in model."""

    on_pre: str = field(default="", init=False)
    model: str


@define(eq=False)
class PreexistingCalcium(CalciumModel):
    """Calcium concentration is pre-existing in neuron model"""

    model: str = field(default="Ca = Ca_pre : mmolar", init=False)


@define(eq=False)
class DynamicCalcium(CalciumModel):
    """Simulates intracellular calcium dynamics from spikes.
    Pieced together from Lütke et al., 2013; Helmchen and Tank, 2015;
    and Song et al., 2021. (`code <https://bitbucket.org/adamshch/naomi_sim/src/25908cf432cd487fffe1f6442548d0fc8f4e8add/code/TimeTraceCode/calcium_dynamics.m?at=master#calcium_dynamics.m>`_)
    """

    on_pre: str = field(default="Ca += dCa_T / (1 + kappa_S + kappa_B)", init=False)
    """from eq 9 in Lütke et al., 2013"""
    model: str = field(
        default="""
            dCa/dt = -gamma * (Ca - Ca_rest) / (1 + kappa_S + kappa_B) : mmolar (clock-driven)
            kappa_B = B_T * K_d / (Ca + K_d)**2 : 1""",
        init=False,
    )
    """from eq 8 in Lütke et al., 2013"""

    Ca_rest: Quantity = field(kw_only=True)
    """resting Ca2+ concentration (molar)"""
    gamma: Quantity = field(kw_only=True)
    """clearance/extrusion rate (1/sec)"""
    B_T: Quantity = field(kw_only=True)
    """total indicator (buffer) concentration (molar)"""
    kappa_S: float = field(kw_only=True)
    """Ca2+ binding ratio of the endogenous buffer"""
    dCa_T: Quantity = field(kw_only=True)
    """total Ca2+ concentration increase per spike (molar)"""

    def init_syn_vars(self, syn: Synapses) -> None:
        syn.Ca = self.Ca_rest


@define(eq=False)
class CalBindingActivationModel:
    """Base class for modeling calcium binding/activation"""

    model: str


@define(eq=False)
class NullBindingActivation(CalBindingActivationModel):
    """Doesn't model binding/activation; i.e., goes straight from [Ca2+] to ΔF/F"""

    model: str = field(default="CaB_active = Ca: mmolar", init=False)


@define(eq=False)
class DoubExpCalBindingActivation(CalBindingActivationModel):
    """Double exponential kernel convolution representing CaB binding/activation.

    Convolution is implemented via ODEs; see ``notebooks/double_exp_conv_as_ode.ipynb``
    for derivation.

    :attr:`A`, :attr:`tau_on`, and :attr:`tau_off` are the versions with proper scale and units of NAOMi's
    ``ca_amp``, ``t_on``, and ``t_off``.

    Some parameters found `here <https://bitbucket.org/adamshch/naomi_sim/src/25908cf432cd487fffe1f6442548d0fc8f4e8add/code/TimeTraceCode/check_cal_params.m?at=master#lines-90>`_.
    Fitting code `here <https://bitbucket.org/adamshch/naomi_sim/src/25908cf432cd487fffe1f6442548d0fc8f4e8add/code/MiscCode/fit_NAOMi_calcium.m?at=master#fit_NAOMi_calcium.m>`_.
    """

    model: str = field(
        default="""
            CaB_active = Ca_rest + b : mmolar  # add tiny bit to avoid /0
            db/dt = beta : mmolar (clock-driven)
            lam = 1/tau_off + 1/tau_on : 1/second
            kap = 1/tau_off : 1/second
            dbeta/dt = (                    # should be M/s/s
                A * (lam - kap) * (Ca - Ca_rest)  # M/s/s
                - (kap + lam) * beta        # M/s/s
                - kap * lam * b    # M/s/s
            ) : mmolar/second (clock-driven)
            """,
        init=False,
    )

    A: float = field(kw_only=True)
    """amplitude of double exponential kernel"""
    tau_on: Quantity = field(kw_only=True)
    """CaB binding/activation time constant (sec)"""
    tau_off: Quantity = field(kw_only=True)
    """CaB unbinding/deactivation time constant (sec)"""
    Ca_rest: Quantity = field(kw_only=True)
    """Resting Ca2+ concentration (molar)."""

    def init_syn_vars(self, syn: Synapses) -> None:
        syn.b = 0
        syn.beta = 0


@define(eq=False)
class ExcitationModel:
    """Defines ``exc_factor``"""

    model: str


@define(eq=False)
class NullExcitation(ExcitationModel):
    """Models excitation as a constant factor"""

    model: str = field(default="exc_factor = 1 : 1", init=False)


@define(eq=False)
class LightExcitation(ExcitationModel):
    """Models light-dependent excitation (not implemented yet)"""

    model: str = field(default="exc_factor = some_function(Irr_pre) : 1", init=False)


@define(eq=False)
class GECI(Sensor):
    """GECI model based on Song et al., 2021, with interchangeable components.

    See :func:`geci` for a convenience function for creating GECI models.

    As a potentially simpler alternative for future work, see the phenomological S2F model
    from `Zhang et al., 2023 <https://www.nature.com/articles/s41586-023-05828-9>`_.
    While parameter count looks similar, at least they have parameters fit already, and directly
    to data, rather than to biophysical processes before the data.
    """

    location: str = field(default="cytoplasm", init=False)

    cal_model: CalciumModel = field(kw_only=True)
    bind_act_model: CalBindingActivationModel = field(kw_only=True)
    exc_model: ExcitationModel = field(kw_only=True)

    fluor_model: str = field(
        default="""
            dFF_baseline = 1 / (1 + (K_d / Ca_rest) ** n_H) : 1
            dFF = exc_factor * rho_rel * dFF_max  * (
                1 / (1 + (K_d / CaB_active) ** n_H)
                - dFF_baseline
            ) : 1
            rho_rel : 1
        """,
        init=False,
    )
    """Uses a Hill equation to convert from Ca2+ to ΔF/F, as in Song et al., 2021"""
    K_d: Quantity = field(kw_only=True)
    """indicator dissociation constant (binding affinity) (molar)"""
    n_H: float = field(kw_only=True)
    """Hill coefficient for conversion from Ca2+ to ΔF/F"""
    dFF_max: float = field(kw_only=True)
    """amplitude of Hill equation for conversion from Ca2+ to ΔF/F,
    Fmax/F0. May be approximated from 'dynamic range' in literature Fmax/Fmin"""

    def get_state(self) -> dict[NeuronGroup, np.ndarray]:
        return {ng_name: syn.dFF for ng_name, syn in self.synapses.items()}

    def __attrs_post_init__(self):
        self.model = "\n".join(
            [
                self.cal_model.model,
                self.bind_act_model.model,
                self.exc_model.model,
                self.fluor_model,
            ]
        )
        self.on_pre = self.cal_model.on_pre

    def init_syn_vars(self, syn: Synapses) -> None:
        for model in [self.cal_model, self.bind_act_model, self.exc_model]:
            if hasattr(model, "init_syn_vars"):
                model.init_syn_vars(syn)

    @property
    def params(self) -> dict:
        """Returns a dictionary of all parameters from model/submodels"""
        params = asdict(self, recurse=False)
        # remove generic fields that are not parameters
        for field in fields_dict(Sensor):
            params.pop(field)
        # remove private attributes
        for key in list(params.keys()):
            if key.startswith("_"):
                params.pop(key)
        # add params from sub-models
        for model in [self.cal_model, self.bind_act_model, self.exc_model]:
            to_add = asdict(model, recurse=False)
            to_add.pop("model")
            params.update(to_add)
        return params


class LightDependentGECI(GECI, LightDependent):
    """Light-dependent calcium indicator (not yet implemented)"""

    pass


def geci(
    light_dependent: bool, doub_exp_conv: bool, pre_existing_cal: bool, **kwparams
) -> GECI:
    """Initializes a :class:`GECI` model with given parameters.

    Parameters
    ----------
    light_dependent : bool
        Whether the indicator is light-dependent.
    doub_exp_conv : bool
        Whether to use double exponential convolution for binding/activation.
    pre_existing_cal : bool
        Whether to use calcium concentrations already simulated in the neuron model.
    **kwparams
        Keyword parameters for :class:`GECI` and sub-models.

    Returns
    -------
    GECI
        A (LightDependent)GECI model specified submodels and parameters.
    """
    ExcModel = LightExcitation if light_dependent else NullExcitation
    CalModel = PreexistingCalcium if pre_existing_cal else DynamicCalcium
    BAModel = DoubExpCalBindingActivation if doub_exp_conv else NullBindingActivation

    def init_from_kwparams(cls, **more_kwargs):
        kwparams.update(more_kwargs)
        kwparams_to_keep = {}
        for field_name, field in fields_dict(cls).items():
            if field.init and field_name in kwparams:
                kwparams_to_keep[field_name] = kwparams[field_name]
        return cls(**kwparams_to_keep)

    GECIClass = LightDependentGECI if light_dependent else GECI

    return init_from_kwparams(
        GECIClass,
        cal_model=init_from_kwparams(CalModel),
        exc_model=init_from_kwparams(ExcModel),
        bind_act_model=init_from_kwparams(BAModel),
        **kwparams,
    )


def _create_geci_fn(
    name,
    K_d,
    n_H,
    dFF_max,
    sigma_noise_rel,
    dFF_1AP_rel=None,
    ca_amp=None,
    t_on=None,
    t_off=None,
    extra_doc="",
):
    """convenience function for creating GECI model functions.

    K_d is in nM.

    dFF_1AP and sigma_noise are relative to GCaMP6s, since noise values obtained indirectly
    from Zhang et al., 2023, supp table 1, which is all relative to GCaMP6s.
    We take ΔF/F / SNR for 1 AP to get a measure of noise.

    We could have done something similar with supp table 1 from Dana 2019,
    but we assume measurements are more accurate from the later paper (as the paper explicitly states).
    Thus, we only use Dana 2019 for the absolute GCaMP6s measurement to scale
    the relative values given in Zhang et al., 2023 and for indiciators not in Zhang 2023
    supp table 1 (GCaMP3).
    """
    gcamp6s_dFF_1AP_dana2019 = 0.133
    gcamp6s_snr_1AP_dana2019 = 4.4
    sigma_gcamp6s = gcamp6s_dFF_1AP_dana2019 / gcamp6s_snr_1AP_dana2019
    sigma_noise = sigma_noise_rel * sigma_gcamp6s
    if dFF_1AP_rel is not None:
        dFF_1AP = dFF_1AP_rel * gcamp6s_dFF_1AP_dana2019
    else:
        dFF_1AP = None

    def geci_fn(
        light_dependent=False,
        doub_exp_conv=True,
        pre_existing_cal=False,
        K_d=K_d * nmolar,
        n_H=n_H,
        dFF_max=dFF_max,
        sigma_noise=sigma_noise,
        dFF_1AP=dFF_1AP,
        ca_amp=ca_amp,
        t_on=t_on,
        t_off=t_off,
        Ca_rest=50 * nmolar,
        kappa_S=110,
        gamma=292.3 / second,
        B_T=200 * umolar,
        dCa_T=7.6 * umolar,  # Lütke et al., 2013 and NAOMi code
        name=name,
        **kwparams,
    ) -> GECI:
        """Returns a (light-dependent) GECI model with specified submodel choices.
        Default parameters are taken from
        `NAOMi's code <https://bitbucket.org/adamshch/naomi_sim/src/25908cf432cd487fffe1f6442548d0fc8f4e8add/code/TimeTraceCode/calcium_dynamics.m?at=master#calcium_dynamics.m>`_
        (Song et al., 2021) as well as Dana et al., 2019 and Zhang et al., 2023.

        Only those parameters used in chosen model components apply.
        If the default is ``None``, then we don't have it fit yet.

        ``ca_amp``, ``t_on``, and ``t_off`` are given as in NAOMi, but are converted to
        the proper scale and units for the double exponential convolution model.
        Namely, ``A = ca_amp / (second / 100)`` and ``tau_[on|off] = second / t_[on|off]``.
        """
        # dt implicit in NAOMi's code, always s/100
        A = ca_amp / (second / 100) if ca_amp else None
        # had to reverse-engineer NAOMi code, which had surprising time constants
        tau_on = second / t_on if t_on else None
        tau_off = second / t_off if t_off else None

        return geci(
            light_dependent,
            doub_exp_conv,
            pre_existing_cal,
            K_d=K_d,
            n_H=n_H,
            dFF_max=dFF_max,
            sigma_noise=sigma_noise,
            dFF_1AP=dFF_1AP,
            A=A,
            tau_on=tau_on,
            tau_off=tau_off,
            Ca_rest=Ca_rest,
            kappa_S=kappa_S,
            gamma=gamma,
            B_T=B_T,
            dCa_T=dCa_T,
            name=name,
            **kwparams,
        )

    geci_fn.__doc__ += "\n\n" + (" " * 8) + extra_doc

    globals()[name] = geci_fn


# from NAOMi:
#                      K_d, n_H, dFF_max, sigma_noise_rel, dFF_1AP_rel, ca_amp, t_on, t_off
_create_geci_fn("gcamp6f", 290, 2.7, 25.2, 1.24, 0.735, 76.1251, 0.8535, 98.6173)
_create_geci_fn("gcamp6s", 147, 2.45, 27.2, 1, 1, 54.6943, 0.4526, 68.5461)
# noise, dFF for GCaMP3 from Dana 2019, since not in Zhang 2023
_create_geci_fn(
    "gcamp3", 287, 2.52, 12, (3.9 / 2.1) / (13.3 / 4.4), 3.9 / 13.3, 0.05, 1, 1
)

# from NAOMi, but
# don't have double exponential convolution parameters for these:
_create_geci_fn(
    "ogb1",
    250,
    1,
    14,
    1,
    extra_doc="*Don't know sigma_noise. Default is the GCaMP6s value.*",
)
_create_geci_fn(
    "gcamp6rs09",
    520,
    3.2,
    25,
    1,
    extra_doc="*Don't know sigma_noise. Default is the GCaMP6s value.*",
)
_create_geci_fn(
    "gcamp6rs06",
    320,
    3,
    15,
    1,
    extra_doc="*Don't know sigma_noise. Default is the GCaMP6s value.*",
)
_create_geci_fn("jgcamp7f", 174, 2.3, 30.2, 0.72, 1.71)
_create_geci_fn("jgcamp7s", 68, 2.49, 40.4, 0.33, 4.96)
_create_geci_fn("jgcamp7b", 82, 3.06, 22.1, 0.25, 4.64)
_create_geci_fn("jgcamp7c", 298, 2.44, 145.6, 0.39, 1.85)



"""Contains basic recorders."""
from typing import Any
from attrs import define, field
from brian2 import (
    PopulationRateMonitor,
    StateMonitor,
    SpikeMonitor,
    Quantity,
    NeuronGroup,
)
import numpy as np
from nptyping import NDArray

from cleo.base import Recorder


@define(eq=False)
class RateRecorder(Recorder):
    """Records firing rate from a single neuron.

    Firing rate comes from Brian's :class:`~brian2.monitors.ratemonitor.PopulationRateMonitor`
    """

    i: int
    """index of neuron to record"""

    mon: PopulationRateMonitor = field(init=False)

    def connect_to_neuron_group(self, neuron_group):
        self.mon = PopulationRateMonitor(neuron_group[self.i])
        self.brian_objects.add(self.mon)

    def get_state(self):
        try:
            return self.mon.rate[-1]
        except IndexError:
            return 0


@define(eq=False)
class VoltageRecorder(Recorder):
    """Records the voltage of a single neuron group."""

    voltage_var_name: str = "v"
    """Name of variable representing membrane voltage"""

    mon: StateMonitor = field(init=False, default=None)

    def connect_to_neuron_group(self, neuron_group):
        if self.mon is not None:
            raise UserWarning(
                "Recorder was already connected to a neuron group. "
                "Can only record from one at a time."
            )
        self.mon = StateMonitor(neuron_group, self.voltage_var_name, record=True)
        self.brian_objects.add(self.mon)

    def get_state(self) -> Quantity:
        """
        Returns
        -------
        Quantity
            Current voltage of target neuron group
        """
        try:
            return self.mon.v[:, -1]
        except IndexError:
            return None


@define(eq=False)
class GroundTruthSpikeRecorder(Recorder):
    """Reports the number of spikes seen since last queried for each neuron.

    This amounts effectively to the number of spikes per control period.
    Note: this will only work for one neuron group at the moment.
    """

    _mon: SpikeMonitor = field(init=False, default=None)

    _num_spikes_seen: int = field(init=False, default=0)

    neuron_group: NeuronGroup = field(init=False, default=None)

    def connect_to_neuron_group(self, neuron_group):
        if self._mon is not None:
            raise UserWarning(
                "Recorder was already connected to a neuron group. "
                "Can only record from one at a time."
            )
        self._mon = SpikeMonitor(neuron_group)
        self.brian_objects.add(self._mon)
        self.neuron_group = neuron_group

    def get_state(self) -> NDArray[(Any,), np.uint]:
        """
        Returns
        -------
        NDArray[(n_neurons,), np.uint]
            n_neurons-length array with spike counts over the latest
            control period.
        """
        num_new_spikes = len(self._mon.t) - self._num_spikes_seen
        self._num_spikes_seen += num_new_spikes
        out = np.zeros(len(self.neuron_group), dtype=np.uint)
        if num_new_spikes > 0:
            for spike_i in self._mon.i[-num_new_spikes:]:
                out[spike_i] += 1
        return out



"""Contains definitions for essential, base classes."""

from __future__ import annotations

import datetime
from abc import ABC, abstractmethod
from typing import Any, Tuple

import neo
from attrs import asdict, define, field, fields_dict
from brian2 import (
    BrianObjectException,
    Equations,
    Network,
    NetworkOperation,
    NeuronGroup,
    Quantity,
    Subgroup,
    Synapses,
    Unit,
    defaultclock,
    ms,
    np,
)
from matplotlib.artist import Artist
from mpl_toolkits.mplot3d import Axes3D

from cleo.registry import registry_for_sim
from cleo.utilities import add_to_neo_segment, analog_signal, brian_safe_name


class NeoExportable(ABC):
    """Mixin class for classes that can be exported to Neo objects"""

    @abstractmethod
    def to_neo(self) -> neo.core.BaseNeo:
        """Return a Neo signal object with the device's data

        Returns
        -------
        neo.core.BaseNeo
            Neo object representing exported data
        """
        pass


@define(eq=False)
class InterfaceDevice(ABC):
    """Base class for devices to be injected into the network"""

    brian_objects: set = field(factory=set, init=False, repr=False)
    """All the Brian objects added to the network by this device.
    Must be kept up-to-date in :meth:`connect_to_neuron_group` and
    other functions so that those objects can be automatically added
    to the network when the device is injected.
    """
    sim: CLSimulator = field(init=False, default=None, repr=False)
    """The simulator the device is injected into """
    name: str = field(kw_only=True)
    """Identifier for device, used in sampling, plotting, etc.
    Name of the class by default. Must be unique among recorders and stimulators"""
    save_history: bool = field(default=True, kw_only=True)
    """Determines whether times and inputs/outputs are recorded.
    True by default.
    
    For stimulators, this is when :meth:`~Stimulator.update` is called.
    For recorders, it is when :meth:`~Recorder.get_state` is called."""

    @name.default
    def _default_name(self) -> str:
        return self.__class__.__name__

    def init_for_simulator(self, simulator: CLSimulator) -> None:
        """Initialize device for simulator on initial injection.

        This function is called only the first time a device is
        injected into a simulator and performs any operations
        that are independent of the individual neuron groups it
        is connected to.

        Parameters
        ----------
        simulator : CLSimulator
            simulator being injected into
        """
        pass

    def reset(self, **kwargs) -> None:
        """Reset the device to a neutral state"""
        pass

    @abstractmethod
    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        """Connect device to given `neuron_group`.

        If your device introduces any objects which Brian must
        keep track of, such as a NeuronGroup, Synapses, or Monitor,
        make sure to add these to :attr:`~cleo.InterfaceDevice.brian_objects`.

        Parameters
        ----------
        neuron_group : NeuronGroup
        **kwparams : optional
            Passed from `inject`
        """
        pass

    def add_self_to_plot(
        self, ax: Axes3D, axis_scale_unit: Unit, **kwargs
    ) -> list[Artist]:
        """Add device to an existing plot

        Should only be called by :func:`~cleo.viz.plot`.

        Parameters
        ----------
        ax : Axes3D
            The existing matplotlib Axes object
        axis_scale_unit : Unit
            The unit used to label axes and define chart limits
        **kwargs : optional

        Returns
        -------
        list[Artist]
            A list of artists used to render the device. Needed for use
            in conjunction with :class:`~cleo.viz.VideoVisualizer`.
        """
        return []

    def update_artists(self, artists: list[Artist], *args, **kwargs) -> list[Artist]:
        """Update the artists used to render the device

        Used to set the artists' state at every frame of a video visualization.
        The current state would be passed in `*args` or `**kwargs`

        Parameters
        ----------
        artists : list[Artist]
            the artists used to render the device originally, i.e.,
            which were returned from the first :meth:`add_self_to_plot` call.

        Returns
        -------
        list[Artist]
            The artists that were actually updated. Needed for efficient
            blit rendering, where only updated artists are re-rendered.
        """
        return []


@define
class IOProcessor(ABC):
    """Abstract class for implementing sampling, signal processing and control

    This must be implemented by the user with their desired closed-loop
    use case, though most users will find the :func:`~processing.LatencyIOProcessor`
    class more useful, since delay handling is already defined.
    """

    sample_period_ms: float = 1
    """Determines how frequently the processor takes samples"""

    latest_ctrl_signal: dict = field(factory=dict, init=False, repr=False)
    """The most recent control signal returned by :meth:`get_ctrl_signals`"""

    @abstractmethod
    def is_sampling_now(self, time) -> bool:
        """Determines whether the processor will take a sample at this timestep.

        Parameters
        ----------
        time : Brian 2 temporal Unit
            Current timestep.

        Returns
        -------
        bool
        """
        pass

    @abstractmethod
    def put_state(self, state_dict: dict, sample_time_ms: float) -> None:
        """Deliver network state to the :class:`IOProcessor`.

        Parameters
        ----------
        state_dict : dict
            A dictionary of recorder measurements, as returned by
            :func:`~cleo.CLSimulator.get_state()`
        sample_time_ms: float
            The current simulation timestep. Essential for simulating
            control latency and for time-varying control.
        """
        pass

    @abstractmethod
    def get_ctrl_signals(self, query_time_ms: float) -> dict:
        """Get per-stimulator control signal from the :class:`~cleo.IOProcessor`.

        Parameters
        ----------
        query_time_ms : float
            Current simulation time.

        Returns
        -------
        dict
            A {'stimulator_name': ctrl_signal} dictionary for updating stimulators.
        """
        pass

    def get_stim_values(self, query_time_ms: float) -> dict:
        ctrl_signals = self.get_ctrl_signals(query_time_ms)
        self.latest_ctrl_signal.update(ctrl_signals)
        stim_value_conversions = self.preprocess_ctrl_signals(
            self.latest_ctrl_signal, query_time_ms
        )
        return ctrl_signals | stim_value_conversions

    def preprocess_ctrl_signals(
        self, latest_ctrl_signals: dict, query_time_ms: float
    ) -> dict:
        """Preprocess control signals as needed to control stimulator waveforms between samples.

        I.e., if a control signal defines the frequency of a periodic light stimulus, this
        function computes the current intensity given the latest frequency and the current
        time. This is called immediately after :meth:`get_ctrl_signals` and on every timestep
        to update the stimulator waveform between samples.

        This only needs to be implemented when a stimulus that varies between samples is desired.
        Otherwise, the control signal returned by :meth:`get_ctrl_signals` is used directly.
        If not all stimulators need this functionality, only return a dict for those that do.
        The original, unprocessed control signal is used for the others.

        Parameters
        ----------
        query_time_ms : float
            Current simulation time.

        Returns
        -------
        dict
            A {'stimulator_name': value} dictionary for updating stimulators.
        """
        return {}

    def get_intersample_ctrl_signal(self, query_time_ms: float) -> dict:
        """Get per-stimulator control signal between samples. I.e., for implementing
        a time-varying waveform based on parameters from the last sample.
        Such parameters will need to be stored in the :class:`~cleo.IOProcessor`."""
        return {}

    def reset(self, **kwargs) -> None:
        pass


@define(eq=False)
class Recorder(InterfaceDevice):
    """Device for taking measurements of the network."""

    @abstractmethod
    def get_state(self) -> Any:
        """Return current measurement."""
        pass


@define(eq=False)
class Stimulator(InterfaceDevice, NeoExportable):
    """Device for manipulating the network"""

    value: Any = field(init=False, default=None)
    """The current value of the stimulator device"""
    default_value: Any = 0
    """The default value of the device---used on initialization and on :meth:`~reset`"""
    t_ms: list[float] = field(factory=list, init=False, repr=False)
    """Times stimulator was updated, stored if :attr:`~cleo.InterfaceDevice.save_history`"""
    values: list[Any] = field(factory=list, init=False, repr=False)
    """Values taken by the stimulator at each :meth:`~update` call, 
    stored if :attr:`~cleo.InterfaceDevice.save_history`"""

    def __attrs_post_init__(self):
        self.value = self.default_value
        self._init_saved_vars()

    def _init_saved_vars(self):
        if self.save_history:
            if self.sim:
                t0 = self.sim.network.t / ms
            else:
                t0 = 0
            self.t_ms = [t0]
            self.values = [self.value]

    def update(self, ctrl_signal) -> None:
        """Set the stimulator value.

        By default this sets :attr:`value` to ``ctrl_signal`` and updates saved
        times and values. You will want to implement this method if your stimulator
        requires additional logic. Use ``super.update(self, value)``
        to preserve the ``self.value`` and ``save_history`` logic

        Parameters
        ----------
        ctrl_signal : any
            The value the stimulator is to take.
        """
        self.value = ctrl_signal
        if self.save_history:
            self.t_ms.append(self.sim.network.t / ms)
            self.values.append(self.value)

    def reset(self, **kwargs) -> None:
        """Reset the stimulator device to a neutral state"""
        self.value = self.default_value
        self._init_saved_vars()

    def to_neo(self):
        signal = analog_signal(self.t_ms, self.values, "dimensionless")
        signal.name = self.name
        signal.description = "Exported from Cleo stimulator device"
        signal.annotate(export_datetime=datetime.datetime.now())
        return signal


@define(eq=False)
class CLSimulator(NeoExportable):
    """The centerpiece of cleo. Integrates simulation components and runs."""

    network: Network = field(repr=False)
    """The Brian network forming the core model"""
    io_processor: IOProcessor = field(default=None, init=False)
    recorders: dict[str, Recorder] = field(factory=dict, init=False, repr=False)
    stimulators: dict[str, Stimulator] = field(factory=dict, init=False, repr=False)
    devices: set[InterfaceDevice] = field(factory=set, init=False)
    _processing_net_op: NetworkOperation = field(default=None, init=False, repr=False)
    _net_store_name: str = field(default="cleo default", init=False, repr=False)

    def inject(
        self, device: InterfaceDevice, *neuron_groups: NeuronGroup, **kwparams: Any
    ) -> CLSimulator:
        """Inject InterfaceDevice into the network, connecting to specified neurons.

        Calls :meth:`~InterfaceDevice.connect_to_neuron_group` for each group with
        kwparams and adds the device's :attr:`~InterfaceDevice.brian_objects`
        to the simulator's :attr:`network`.

        Parameters
        ----------
        device : InterfaceDevice
            Device to inject

        Returns
        -------
        CLSimulator
            self
        """
        if len(neuron_groups) == 0:
            raise Exception("Injecting stimulator for no neuron groups is meaningless.")
        for ng in neuron_groups:
            if type(ng) == NeuronGroup:
                if ng not in self.network.objects:
                    raise Exception(
                        f"Attempted to connect device {device.name} to neuron group "
                        f"{ng.name}, which is not part of the simulator's network."
                    )
            elif type(ng) == Subgroup:
                # must look at sorted_objects because ng.source is unhashable
                if ng.source not in self.network.sorted_objects:
                    raise Exception(
                        f"Attempted to connect device {device.name} to neuron group "
                        f"{ng.source.name}, which is not part of the simulator's network."
                    )
            if device.sim not in [None, self]:
                raise Exception(
                    f"Attempted to inject device {device.name} into {self}, "
                    f"but it was previously injected into {device.sim}. "
                    "Each device can only be injected into one CLSimulator."
                )
            if device.sim is None:
                device.sim = self
                device.init_for_simulator(self)
            device.connect_to_neuron_group(ng, **kwparams)
        for brian_object in device.brian_objects:
            if brian_object not in self.network.objects:
                self.network.add(brian_object)
        self.network.store(self._net_store_name)
        if isinstance(device, Recorder):
            if (
                device.name in self.recorders
                and device is not self.recorders[device.name]
            ):
                raise ValueError(
                    f"Another Recorder with name {device.name} has already been injected"
                )
            self.recorders[device.name] = device
        if isinstance(device, Stimulator):
            if (
                device.name in self.stimulators
                and device is not self.stimulators[device.name]
            ):
                raise ValueError(
                    f"Another Stimulator with name {device.name} has already been injected"
                )
            self.stimulators[device.name] = device
        self.devices.add(device)
        return self

    def get_state(self) -> dict:
        """Return current recorder measurements.

        Returns
        -------
        dict
            A dictionary of `name`: `state` pairs for
            all recorders in the simulator.
        """
        state = {}
        for name, recorder in self.recorders.items():
            state[name] = recorder.get_state()
        return state

    def update_stimulators(self, stim_values: dict[str, Any]) -> None:
        """Update stimulators with output from the :class:`IOProcessor`

        Parameters
        ----------
        stim_values : dict
            {`stimulator_name`: `stim_value`} dictionary with values
            to update each stimulator.
        """
        for name, value in stim_values.items():
            self.stimulators[name].update(value)

    def set_io_processor(
        self, io_processor: IOProcessor, communication_period=None
    ) -> CLSimulator:
        """Set simulator IO processor

        Will replace any previous IOProcessor so there is only one at a time.
        A Brian NetworkOperation is created to govern communication between
        the Network and the IOProcessor.

        Parameters
        ----------
        io_processor : IOProcessor

        Returns
        -------
        CLSimulator
            self
        """
        self.io_processor = io_processor
        # remove previous NetworkOperation
        if self._processing_net_op is not None:
            self.network.remove(self._processing_net_op)
            self._processing_net_op = None

        if io_processor is None:
            return

        def communicate_with_io_proc(t):
            if io_processor.is_sampling_now(t / ms):
                io_processor.put_state(self.get_state(), t / ms)
            stim_values = io_processor.get_stim_values(t / ms)
            self.update_stimulators(stim_values)

        # communication should be at every timestep. The IOProcessor
        # decides when to sample and deliver results.
        if communication_period is None:
            communication_period = defaultclock.dt
        self._processing_net_op = NetworkOperation(
            communicate_with_io_proc, dt=communication_period
        )
        self.network.add(self._processing_net_op)
        self.network.store(self._net_store_name)
        return self

    def run(self, duration: Quantity, **kwparams) -> None:
        """Run simulation.

        Parameters
        ----------
        duration : brian2 temporal Quantity
            Length of simulation
        **kwparams : additional arguments passed to brian2.run()
            level has a default value of 1
        """
        level = kwparams.get("level", 1)
        kwparams["level"] = level
        self.network.run(duration, **kwparams)

    def reset(self, **kwargs):
        """Reset the simulator to a neutral state

        Restores the Brian Network to where it was when the
        CLSimulator was last modified (last injection, IOProcessor change).
        Calls reset() on :attr:`devices` and :attr:`io_processor`.
        """
        # kwargs passed to stimulators, recorders, and io_processor reset
        self.network.restore(self._net_store_name)
        for device in self.devices:
            device.reset(**kwargs)
        if self.io_processor is not None:
            self.io_processor.reset(**kwargs)

    def to_neo(self) -> neo.core.Block:
        """Exports simulator data to a Neo Block

        Returns
        -------
        neo.core.Block
            Neo Block containing signals representing each device's data
        """
        block = neo.Block(
            description="Exported from Cleo simulation",
        )
        block.annotate(export_datetime=datetime.datetime.now())
        seg = neo.Segment()
        block.segments.append(seg)
        for device in self.devices:
            if not isinstance(device, NeoExportable):
                continue
            dev_neo = device.to_neo()
            if isinstance(dev_neo, neo.core.Group):
                data_objects = dev_neo.data_children_recur
                block.groups.append(dev_neo)
            elif isinstance(dev_neo, neo.core.dataobject.DataObject):
                data_objects = [dev_neo]
            add_to_neo_segment(seg, *data_objects)
        return block


@define(eq=False)
class SynapseDevice(InterfaceDevice):
    """Base class for devices that record from/stimulate neurons via
    a Synapses object with device-specific model. Used for opsin and
    indicator classes"""

    model: str = field(init=False)
    """Basic Brian model equations string.
    
    Should contain a `rho_rel` term reflecting relative expression 
    levels. Will likely also contain special NeuronGroup-dependent
    symbols such as V_VAR_NAME to be replaced on injection in 
    :meth:`modify_model_and_params_for_ng`."""

    on_pre: str = field(init=False, default="")
    """Model string for :class:`brian2.synapses.synapses.Synapses` reacting to spikes."""

    synapses: dict[str, Synapses] = field(factory=dict, init=False, repr=False)
    """Stores the synapse objects implementing the model, connecting from source
    (light aggregator neurons or the target group itself) to target neuron groups,
    of form ``{target_ng.name: synapses}``."""

    source_ngs: dict[str, NeuronGroup] = field(factory=dict, init=False, repr=False)
    """``{target_ng.name: source_ng}`` dict of source neuron groups.
    
    The source is the target itself by default or light aggregator neurons for
    :class:`~cleo.light.LightDependent`."""

    per_ng_unit_replacements: list[Tuple[str, str]] = field(
        factory=list, init=False, repr=False
    )
    """List of (UNIT_NAME, neuron_group_specific_unit_name) tuples to be substituted
    in the model string on injection and before checking required variables."""

    required_vars: list[Tuple[str, Unit]] = field(factory=list, init=False, repr=False)
    """Default names of state variables required in the neuron group,
    along with units, e.g., [('Iopto', amp)].
    
    It is assumed that non-default values can be passed in on injection
    as a keyword argument ``[default_name]_var_name=[non_default_name]``
    and that these are found in the model string as 
    ``[DEFAULT_NAME]_VAR_NAME`` before replacement."""

    extra_namespace: dict = field(factory=dict, repr=False)
    """Additional items (beyond parameters) to be added to the opto synapse namespace"""

    def _get_source_for_synapse(
        self, target_ng: NeuronGroup, i_targets: list[int]
    ) -> Tuple[NeuronGroup, list[int]]:
        """Get the source neuron group and indices of source neurons.

        Parameters
        ----------
        ng : NeuronGroup
            The target neuron group.
        i_targets : list[int]
            The indices of the target neurons in the target neuron group.

        Returns
        -------
        Tuple[NeuronGroup, list[int]]
            A tuple containing the source neuron group and indices to use in Synapses
        """
        # by default the source is the target group itself
        return target_ng, i_targets

    def connect_to_neuron_group(self, neuron_group: NeuronGroup, **kwparams) -> None:
        """Transfect neuron group with device.

        Parameters
        ----------
        neuron_group : NeuronGroup
            The neuron group to transform

        Keyword args
        ------------
        p_expression : float
            Probability (0 <= p <= 1) that a given neuron in the group
            will express the protein. 1 by default.
        i_targets : array-like
            Indices of neurons in the group to transfect. recommended for efficiency
            when stimulating or imaging a small subset of the group.
            Incompatible with ``p_expression``.
        rho_rel : float
            The expression level, relative to the standard model fit,
            of the protein. 1 by default. For heterogeneous expression,
            this would have to be modified in the light-dependent synapse
            post-injection, e.g., ``opsin.syns["neuron_group_name"].rho_rel = ...``
        [default_name]_var_name : str
            See :attr:`~required_vars`. Allows for custom variable names.
        """
        if neuron_group.name in self.source_ngs:
            assert neuron_group.name in self.synapses
            raise ValueError(
                f"{self.__class__.__name__} {self.name} already connected to neuron group"
                f" {neuron_group.name}"
            )

        # get modified synapse model string (i.e., with names/units specified)
        mod_syn_model, mod_syn_params = self.modify_model_and_params_for_ng(
            neuron_group, kwparams
        )

        # handle p_expression
        if "p_expression" in kwparams:
            if "i_targets" in kwparams:
                raise ValueError("p_expression and i_targets are incompatible")
            p_expression = kwparams.get("p_expression", 1)
            expr_bool = np.random.rand(neuron_group.N) < p_expression
            i_targets = np.where(expr_bool)[0]
        elif "i_targets" in kwparams:
            i_targets = kwparams["i_targets"]
        else:
            i_targets = list(range(neuron_group.N))
        if len(i_targets) == 0:
            return

        source_ng, i_sources = self._get_source_for_synapse(neuron_group, i_targets)

        syn = Synapses(
            source_ng,
            neuron_group,
            model=mod_syn_model,
            on_pre=self.on_pre,
            namespace=mod_syn_params,
            name=f"syn_{brian_safe_name(self.name)}_{neuron_group.name}",
        )
        syn.namespace.update(self.extra_namespace)
        syn.connect(i=i_sources, j=i_targets)
        self.init_syn_vars(syn)
        # relative protein density
        syn.rho_rel = kwparams.get("rho_rel", 1)

        # store at the end, after all checks have passed
        self.source_ngs[neuron_group.name] = source_ng
        self.brian_objects.add(source_ng)
        self.synapses[neuron_group.name] = syn
        self.brian_objects.add(syn)

        registry = registry_for_sim(self.sim)
        registry.register(self, neuron_group)

    def modify_model_and_params_for_ng(
        self, neuron_group: NeuronGroup, injct_params: dict
    ) -> Tuple[Equations, dict]:
        """Adapt model for given neuron group on injection

        This enables the specification of variable names
        differently for each neuron group, allowing for custom names
        and avoiding conflicts.

        Parameters
        ----------
        neuron_group : NeuronGroup
            NeuronGroup this opsin model is being connected to
        injct_params : dict
            kwargs passed in on injection, could contain variable
            names to plug into the model

        Keyword Args
        ------------
        model : str, optional
            Model to start with, by default that defined for the class.
            This allows for prior string manipulations before it can be
            parsed as an `Equations` object.

        Returns
        -------
        Equations, dict
            A tuple containing an Equations object
            and a parameter dictionary, constructed from :attr:`~model`
            and :attr:`~params`, respectively, with modified names for use
            in :attr:`~cleo.opto.OptogeneticIntervention.synapses`
        """
        model = self.model

        # perform unit substitutions
        for unit_name, neuron_group_unit_name in self.per_ng_unit_replacements:
            model = model.replace(unit_name, neuron_group_unit_name)

        # check required variables/units and replace placeholder names
        for default_name, unit in self.required_vars:
            var_name = injct_params.get(f"{default_name}_var_name", default_name)
            if var_name not in neuron_group.variables or not neuron_group.variables[
                var_name
            ].unit.has_same_dimensions(unit):
                raise BrianObjectException(
                    (
                        f"{var_name} : {unit.name} needed in the model of NeuronGroup "
                        f"{neuron_group.name} to connect Opsin {self.name}."
                    ),
                    neuron_group,
                )
            # opsin synapse model needs modified names
            to_replace = f"{default_name}_var_name".upper()
            model = model.replace(to_replace, var_name)

        # Synapse variable and parameter names cannot be the same as any
        # neuron group variable name
        return self._fix_name_conflicts(model, neuron_group)

    @property
    def params(self) -> dict:
        """Returns a dictionary of parameters for the model"""
        params = asdict(self, recurse=False)
        # remove generic fields that are not parameters
        # assume we only want fields in the last class in the
        # hierarchy
        parent_class = type(self).__mro__[1]
        for field in fields_dict(parent_class):
            params.pop(field)
        # remove private attributes
        for key in list(params.keys()):
            if key.startswith("_"):
                params.pop(key)
        return params

    def _fix_name_conflicts(
        self, modified_model: str, neuron_group: NeuronGroup
    ) -> Tuple[Equations, dict]:
        modified_params = self.params.copy()
        rename = lambda x: f"{x}_syn"

        # get variables to rename
        opsin_eqs = Equations(modified_model)
        substitutions = {}
        for var in opsin_eqs.names:
            if var in neuron_group.variables:
                substitutions[var] = rename(var)

        # and parameters
        for param in self.params.keys():
            if param in neuron_group.variables:
                substitutions[param] = rename(param)
                modified_params[rename(param)] = modified_params[param]
                del modified_params[param]

        mod_opsin_eqs = opsin_eqs.substitute(**substitutions)
        return mod_opsin_eqs, modified_params

    def reset(self, **kwargs):
        for opto_syn in self.synapses.values():
            self.init_syn_vars(opto_syn)

    def init_syn_vars(self, syn: Synapses) -> None:
        """Initializes appropriate variables in Synapses implementing the model

        Also called on :meth:`reset`.

        Parameters
        ----------
        syn : Synapses
            The synapses object implementing this model
        """
        pass



from typing import Any

from attrs import define
from brian2 import Quantity, np
from brian2.units import nmeter, um
from nptyping import NDArray

from cleo.coords import concat_coords, coords_from_ng, coords_from_xyz
from cleo.light import Light, LightModel
from cleo.utilities import uniform_cylinder_rθz, xyz_from_rθz


@define(eq=False)
class GaussianEllipsoid(LightModel):
    sigma_axial: Quantity = 18 * um
    """Standard deviation distance along the focal axis.

    Standard deviations estimated by taking point where response is ~60% of peak:
    
    ======================== ========== ============= =======
    Publication              axial      lateral       measure
    ======================== ========== ============= =======
    Prakash et al., 2012     13 μm      7 μm          photocurrent
    Rickgauer et al., 2014   8 μm       4 μm          Ca2+ dF/F response
    Packer et al., 2015      18 μm      8 μm          AP probability
    Chen et al., 2019        18/11 μm   8/? μm        AP probability/photocurrent
    ======================== ========== ============= =======
    """
    sigma_lateral: Quantity = 8 * um
    """Standard deviation distance along the focal plane."""

    def transmittance(
        self,
        source_coords: Quantity,
        source_dir_uvec: NDArray[(Any, 3), Any],
        target_coords: Quantity,
    ) -> NDArray[(Any, Any), float]:
        assert np.allclose(np.linalg.norm(source_dir_uvec, axis=-1), 1)
        r, z = self._get_rz_for_xyz(source_coords, source_dir_uvec, target_coords)
        return self._gaussian_transmittance(r, z)

    def viz_params(
        self,
        coords: Quantity,
        direction: NDArray[(Any, 3), Any],
        T_threshold: float,
        n_points_per_source: int = 4000,
        **kwargs,
    ) -> Quantity:
        r_thresh, zc_thresh = self._find_rz_thresholds(T_threshold)
        r, theta, zc = uniform_cylinder_rθz(
            n_points_per_source, r_thresh, zc_thresh * 2
        )
        zc -= zc_thresh

        end = coords + zc_thresh * direction
        x, y, z = xyz_from_rθz(r, theta, zc, coords, end)
        # m x n x 3
        density_factor = 3
        cyl_vol = np.pi * r_thresh**2 * zc_thresh
        markersize_um = (cyl_vol / n_points_per_source * density_factor) ** (1 / 3) / um
        intensity_scale = (1000 / n_points_per_source) ** (1 / 3)
        return coords_from_xyz(x, y, z), markersize_um, intensity_scale

    def _gaussian_transmittance(self, r, z):
        """r is lateral distance, z is axial distance from focal point.

        Not normalizing the Gaussian: i.e., a point perfectly in focus will
        get relative power of 1.0, not 1.0 / (2 * pi * sigma**2)."""
        return np.exp(-(r**2) / (2 * self.sigma_lateral**2)) * np.exp(
            -(z**2) / (2 * self.sigma_axial**2)
        )

    def _find_rz_thresholds(self, thresh):
        """find r and z thresholds for visualization purposes"""
        res_um = 0.1
        zc = np.arange(1000, 0, -res_um) * um  # ascending T
        T = self._gaussian_transmittance(0 * um, zc)
        try:
            zc_thresh = zc[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            zc_thresh = np.max(zc)

        r = np.arange(100, 0, -res_um) * um
        T = self._gaussian_transmittance(r, 0 * um)
        try:
            r_thresh = r[np.searchsorted(T, thresh)]
        except IndexError:  # no points above threshold
            r_thresh = np.max(r)

        return r_thresh, zc_thresh


def tp_light_from_scope(scope, wavelength=1060 * nmeter, **kwargs) -> Light:
    """Creates a light object from a scope object with 2P focused laser points
    at each target.

    Parameters
    ----------
    scope : Scope
        The scope object containing the laser spots.
    wavelength : Quantity, optional
        The wavelength of the laser, by default 1060 * nmeter.
    """
    coords = []
    for ng, i_targets in zip(scope.neuron_groups, scope.i_targets_per_injct):
        coords.append(coords_from_ng(ng)[i_targets])
    coords = concat_coords(*coords)
    light = Light(
        coords=coords,
        direction=scope.direction,
        light_model=GaussianEllipsoid(),
        wavelength=wavelength,
        **kwargs,
    )
    return light



"""Contains opsin models and default parameters"""
from __future__ import annotations

from typing import Any, Tuple

from attrs import define, field
from brian2 import (
    Synapses,
    Unit,
    check_units,
    get_unit,
    implementation,
    np,
)
from brian2.units import (
    Quantity,
    amp,
    cm2,
    mM,
    mm2,
    ms,
    msiemens,
    mV,
    nsiemens,
    pcoulomb,
    psiemens,
    second,
    umeter,
    volt,
)
from brian2.units.allunits import radian

from cleo.base import SynapseDevice
from cleo.light import LightDependent


# hacky MRO stuff: base class order is important here
@define(eq=False)
class Opsin(LightDependent, SynapseDevice):
    """Base class for opsin model.

    Requires that the neuron model has a current term
    (by default Iopto) which is assumed to be positive (unlike the
    convention in many opsin modeling papers, where the current is
    described as negative).

    We approximate dynamics under multiple wavelengths using a weighted sum
    of photon fluxes, where the ε factor indicates the activation
    relative to the peak-sensitivy wavelength for an equivalent number of photons
    (see Mager et al, 2018). This weighted sum is an approximation of a nonlinear
    peak-non-peak wavelength relation; see ``notebooks/multi_wavelength_model.ipynb``
    for details."""

    @property
    def action_spectrum(self):
        """Alias for ``light_receptor.spectrum``"""
        return self.spectrum


@define(eq=False)
class MarkovOpsin(Opsin):
    """Base class for Markov state models à la Evans et al., 2016"""

    required_vars: list[Tuple[str, Unit]] = field(
        factory=lambda: [("Iopto", amp), ("v", volt)],
        init=False,
    )


@implementation(
    "cython",
    """
    cdef double f_unless_x0(double f, double x, double f_when_x0):
        if x == 0:
            return f_when_x0
        else:
            return f
    """,
)
@check_units(f=1, x=volt, f_when_0=1, result=1)
def f_unless_x0(f, x, f_when_x0):
    f[x == 0] = f_when_x0
    return f


@define(eq=False)
class FourStateOpsin(MarkovOpsin):
    """4-state model from PyRhO (Evans et al. 2016).

    rho_rel is channel density relative to standard model fit;
    modifying it post-injection allows for heterogeneous opsin expression.

    IOPTO_VAR_NAME and V_VAR_NAME are substituted on injection.

    Defaults are for ChR2.
    """

    g0: Quantity = 114000 * psiemens
    gamma: Quantity = 0.00742
    phim: Quantity = 2.33e17 / mm2 / second  # *photon, not in Brian2
    k1: Quantity = 4.15 / ms
    k2: Quantity = 0.868 / ms
    p: Quantity = 0.833
    Gf0: Quantity = 0.0373 / ms
    kf: Quantity = 0.0581 / ms
    Gb0: Quantity = 0.0161 / ms
    kb: Quantity = 0.063 / ms
    q: Quantity = 1.94
    Gd1: Quantity = 0.105 / ms
    Gd2: Quantity = 0.0138 / ms
    Gr0: Quantity = 0.00033 / ms
    E: Quantity = 0 * mV
    v0: Quantity = 43 * mV
    v1: Quantity = 17.1 * mV
    model: str = field(
        init=False,
        default="""
        dC1/dt = Gd1*O1 + Gr0*C2 - Ga1*C1 : 1 (clock-driven)
        dO1/dt = Ga1*C1 + Gb*O2 - (Gd1+Gf)*O1 : 1 (clock-driven)
        dO2/dt = Ga2*C2 + Gf*O1 - (Gd2+Gb)*O2 : 1 (clock-driven)
        C2 = 1 - C1 - O1 - O2 : 1

        Theta = int(phi_pre > 0*phi_pre) : 1
        Hp = Theta * phi_pre**p/(phi_pre**p + phim**p) : 1
        Ga1 = k1*Hp : hertz
        Ga2 = k2*Hp : hertz
        Hq = Theta * phi_pre**q/(phi_pre**q + phim**q) : 1
        Gf = kf*Hq + Gf0 : hertz
        Gb = kb*Hq + Gb0 : hertz

        fphi = O1 + gamma*O2 : 1
        # v1/v0 when v-E == 0 via l'Hopital's rule
        fv = f_unless_x0(
            (1 - exp(-(V_VAR_NAME_post-E)/v0)) / ((V_VAR_NAME_post-E)/v1),
            V_VAR_NAME_post - E,
            v1/v0
        ) : 1

        IOPTO_VAR_NAME_post = -g0*fphi*fv*(V_VAR_NAME_post-E)*rho_rel : ampere (summed)
        rho_rel : 1""",
    )

    extra_namespace: dict[str, Any] = field(
        init=False, factory=lambda: {"f_unless_x0": f_unless_x0}
    )

    def init_syn_vars(self, opto_syn: Synapses) -> None:
        for varname, value in {"C1": 1, "O1": 0, "O2": 0}.items():
            setattr(opto_syn, varname, value)


@define(eq=False)
class BansalFourStateOpsin(MarkovOpsin):
    """4-state model from Bansal et al. 2020.

    The difference from the PyRhO model is that there is no voltage dependence.

    rho_rel is channel density relative to standard model fit;
    modifying it post-injection allows for heterogeneous opsin expression.

    IOPTO_VAR_NAME and V_VAR_NAME are substituted on injection.
    """

    Gd1: Quantity = 0.066 / ms
    Gd2: Quantity = 0.01 / ms
    Gr0: Quantity = 3.33e-4 / ms
    g0: Quantity = 3.2 * nsiemens
    phim: Quantity = 1e16 / mm2 / second  # *photon, not in Brian2
    k1: Quantity = 0.4 / ms
    k2: Quantity = 0.12 / ms
    Gf0: Quantity = 0.018 / ms
    Gb0: Quantity = 0.008 / ms
    kf: Quantity = 0.01 / ms
    kb: Quantity = 0.008 / ms
    gamma: Quantity = 0.05
    p: Quantity = 1
    q: Quantity = 1
    E: Quantity = 0 * mV

    model: str = field(
        init=False,
        default="""
        dC1/dt = Gd1*O1 + Gr0*C2 - Ga1*C1 : 1 (clock-driven)
        dO1/dt = Ga1*C1 + Gb*O2 - (Gd1+Gf)*O1 : 1 (clock-driven)
        dO2/dt = Ga2*C2 + Gf*O1 - (Gd2+Gb)*O2 : 1 (clock-driven)
        C2 = 1 - C1 - O1 - O2 : 1

        Theta = int(phi_pre > 0*phi_pre) : 1
        Hp = Theta * phi_pre**p/(phi_pre**p + phim**p) : 1
        Ga1 = k1*Hp : hertz
        Ga2 = k2*Hp : hertz
        Hq = Theta * phi_pre**q/(phi_pre**q + phim**q) : 1
        Gf = kf*Hq + Gf0 : hertz
        Gb = kb*Hq + Gb0 : hertz

        fphi = O1 + gamma*O2 : 1

        IOPTO_VAR_NAME_post = -g0*fphi*(V_VAR_NAME_post-E)*rho_rel : ampere (summed)
        rho_rel : 1""",
    )

    def init_syn_vars(self, opto_syn: Synapses) -> None:
        for varname, value in {"C1": 1, "O1": 0, "O2": 0}.items():
            setattr(opto_syn, varname, value)


@define(eq=False)
class BansalThreeStatePump(MarkovOpsin):
    """3-state model from `Bansal et al. 2020 <https://iopscience.iop.org/article/10.1088/2057-1976/ab90a1>`_.
    Defaults are for eNpHR3.0.

    rho_rel is channel density relative to standard model fit;
    modifying it post-injection allows for heterogeneous opsin expression.

    IOPTO_VAR_NAME and V_VAR_NAME are substituted on injection.
    """

    Gd: Quantity = 0.25 / ms
    Gr: Quantity = 0.05 / ms
    ka: Quantity = 1 / ms
    p: Quantity = 0.7
    q: Quantity = 0.1
    phim: Quantity = 1.2e18 / mm2 / second  # photons
    E: Quantity = -400 * mV
    g0: Quantity = 22.34 * nsiemens
    a: Quantity = 0.02e-2 * mM / pcoulomb
    b: float = 12
    vartheta_max: Quantity = 5 * mM / second
    kd: Quantity = 16 * mM
    # Sukhdev Roy said they used 2.3 nS for the 'photocurrent simulation' and 2.3 mS/cm² for the 'neuronal simulation', whatever that means
    # g_Cl: Quantity = 2.3 * msiemens / cm2 * (30090 * umeter**2)  # surface area
    g_Cl: Quantity = 2.3 * nsiemens
    Cl_out: Quantity = 124 * mM
    Psi0: Quantity = 4.4286 * mM / second
    E_Cl0: Quantity = -70 * mV
    vmin: Quantity = -400 * mV
    """Needed to avoid jumps in [Cl_in] for EIF neurons"""
    vmax: Quantity = 50 * mV
    """Needed to avoid jumps in [Cl_in] for EIF neurons"""
    model: str = field(
        init=False,
        default="""
        dP0/dt = Gr*P6 - Ga*P0 : 1 (clock-driven)
        dP4/dt = Ga*P0 - Gd*P4 : 1 (clock-driven)
        P6 = 1 - P0 - P4 : 1

        Theta = int(phi_pre > 0*phi_pre) : 1
        Hp = Theta * phi_pre**p/(phi_pre**p + phim**p) : 1
        Ga = ka*Hp : hertz

        fphi = P4 : 1
        dCl_in/dt = a*(I_i + b*I_Cl_leak) : mmolar (clock-driven)
        E_Cl = -26.67*mV * log(Cl_out/Cl_in) : volt
        I_Cl_leak = g_Cl * (E_Cl0 - E_Cl) : ampere

        Psi = vartheta_max*Cl_out / (kd + Cl_out) / Psi0 : 1
        I_i = g0*fphi*Psi*rho_rel * (clip(V_VAR_NAME_post, vmin, vmax)-E) : ampere

        IOPTO_VAR_NAME_post = -(I_i + I_Cl_leak) : ampere (summed)
        rho_rel : 1""",
    )

    def init_syn_vars(self, opto_syn: Synapses) -> None:
        opto_syn.P0 = 1
        opto_syn.P4 = 0
        # P6 automatically set since it's 1 - P0 - P4
        RToverF = 26.67 * mV
        # need to remove and add back on units for log/exp to work
        opto_syn.Cl_in = np.exp(self.E_Cl0 / RToverF + np.log(self.Cl_out / mM)) * mM


@define(eq=False)
class ProportionalCurrentOpsin(Opsin):
    """A simple model delivering current proportional to light intensity"""

    I_per_Irr: Quantity = field(kw_only=True)
    """ How much current (in amps or unitless, depending on neuron model)
    to deliver per mW/mm2.
    """
    # would be IOPTO_UNIT but that throws off Equation parsing
    model: str = field(
        init=False,
        default="""
            IOPTO_VAR_NAME_post = I_per_Irr / (mwatt / mm2) 
                * Irr_pre * rho_rel : IOPTO_UNIT (summed)
            rho_rel : 1
        """,
    )

    required_vars: list[Tuple[str, Unit]] = field(factory=list, init=False)

    def __attrs_post_init__(self):
        if isinstance(self.I_per_Irr, Quantity):
            Iopto_unit = get_unit(self.I_per_Irr.dim)
        else:
            Iopto_unit = radian
        self.per_ng_unit_replacements = [("IOPTO_UNIT", Iopto_unit.name)]
        self.required_vars = [("Iopto", Iopto_unit)]



"""Contains opsin models, light sources, and some parameters"""
from cleo.opto.opsin_library import (
    chr2_4s,
    chr2_b4s,
    chr2_h134r_4s,
    chrimson_4s,
    enphr3_3s,
    gtacr2_4s,
    vfchrimson_4s,
)
from cleo.opto.opsins import (
    BansalFourStateOpsin,
    BansalThreeStatePump,
    FourStateOpsin,
    Opsin,
    ProportionalCurrentOpsin,
)



"""Code for orchestrating inter-device interactions. 

This should only be relevant for developers, not users, as this code is used
under the hood when interacting devices are injected (e.g., light and opsin)."""
from __future__ import annotations

from typing import Tuple

from attrs import define, field
from brian2 import NeuronGroup, Subgroup, Synapses
from brian2.units.allunits import joule, kgram, meter, meter2, nmeter, second

from cleo.coords import coords_from_ng
from cleo.utilities import brian_safe_name


@define(repr=False)
class DeviceInteractionRegistry:
    """Facilitates the creation and maintenance of 'neurons' and 'synapses'
    implementing many-to-many light-opsin/indicator relationships"""

    sim: "CLSimulator" = field()

    subgroup_idx_for_light: dict["Light", slice] = field(factory=dict, init=False)
    """Maps light to its indices in the :attr:`light_source_ng`"""

    light_source_ng: NeuronGroup = field(init=False, default=None)
    """Represents ALL light sources (multiple devices)"""

    ldds_for_ng: dict[NeuronGroup, set["LightDependent"]] = field(
        factory=dict, init=False
    )
    """Maps neuron group to the light-dependent devices injected into it"""

    lights_for_ng: dict[NeuronGroup, set["Light"]] = field(factory=dict, init=False)
    """Maps neuron group to the lights injected into it"""

    light_prop_syns: dict[Tuple["LightDependent", NeuronGroup], Synapses] = field(
        factory=dict, init=False
    )
    """Maps (light-dependent device, neuron group) to the synapses implementing light propagation"""

    connections: set[Tuple["Light", "LightDependent", NeuronGroup]] = field(
        factory=set, init=False
    )
    """Set of (light, light-dependent device, neuron group) tuples representing
    previously created connections."""

    light_prop_model = """
        T : 1
        epsilon : 1
        Ephoton : joule
        Irr_post = epsilon * T * Irr0_pre : watt/meter**2 (summed)
        phi_post = Irr_post / Ephoton : 1/second/meter**2 (summed)
    """
    """Model used in light propagation synapses"""

    brian_objects: set = field(factory=set, init=False)
    """Stores all Brian objects created (and injected into the network) by this registry"""

    def register(self, device: "InterfaceDevice", ng: NeuronGroup) -> None:
        """Registers a device injection with the registry.

        Parameters
        ----------
        device : InterfaceDevice
            Device being injected
        ng : NeuronGroup
            Neurons being injected into
        """
        ancestor_classes = [t.__name__ for t in type(device).__mro__]
        if "Light" in ancestor_classes:
            self.register_light(device, ng)
        elif "LightDependent" in ancestor_classes:
            self.register_ldd(device, ng)

    def connect_light_to_ldd_for_ng(
        self, light: "Light", ldd: "LightDependent", ng: NeuronGroup
    ) -> None:
        """Connects a light to a light-dependent device for a given neuron group.

        Parameters
        ----------
        light : Light
            Light being injected
        ldd : LightDependent
            Light-dependent device the light will affect
        ng : NeuronGroup
            Neurons affected by the light-dependent device

        Raises
        ------
        ValueError
            if the connection has already been made
        """
        epsilon = ldd.epsilon(light.wavelength / nmeter)
        if epsilon == 0:
            return

        light_prop_syn = self._get_or_create_light_prop_syn(ldd, ng)
        if (light, ldd, ng) in self.connections:
            raise ValueError(f"{light} already connected to {ldd.name} for {ng.name}")

        i_source = self.subgroup_idx_for_light[light]
        light_prop_syn.epsilon[i_source, :] = epsilon
        light_prop_syn.T[i_source, :] = light.transmittance(coords_from_ng(ng)).ravel()
        # fmt: off
        # Ephoton = h*c/lambda
        light_prop_syn.Ephoton[i_source, :] = (
            6.63e-34 * meter2 * kgram / second
            * 2.998e8 * meter / second
            / light.wavelength
        )
        # fmt: on
        self.connections.add((light, ldd, ng))

    def _add_brian_object(self, obj):
        self.brian_objects.add(obj)
        self.sim.network.add(obj)

    def _remove_brian_object(self, obj):
        self.brian_objects.remove(obj)
        self.sim.network.remove(obj)

    def _get_or_create_light_prop_syn(
        self, ldd: "LightDependent", ng: NeuronGroup
    ) -> Synapses:
        if (ldd, ng) not in self.light_prop_syns:
            light_agg_ng = ldd.light_agg_ngs[ng.name]

            light_prop_syn = Synapses(
                self.light_source_ng,
                light_agg_ng,
                model=self.light_prop_model,
                name=f"light_prop_{brian_safe_name(ldd.name)}_{ng.name}",
            )
            light_prop_syn.connect()
            # non-zero initialization to avoid nans from /0
            light_prop_syn.Ephoton = 1 * joule
            self._add_brian_object(light_prop_syn)
            self.light_prop_syns[(ldd, ng)] = light_prop_syn

        return self.light_prop_syns[(ldd, ng)]

    def register_ldd(self, ldd: "LightDependent", ng: NeuronGroup):
        """Connects lights previously injected into this neuron group to this light-dependent device"""
        if ng not in self.ldds_for_ng:
            self.ldds_for_ng[ng] = set()
        self.ldds_for_ng[ng].add(ldd)
        prev_injct_lights = self.lights_for_ng.get(ng, set())
        for light in prev_injct_lights:
            self.connect_light_to_ldd_for_ng(light, ldd, ng)

    def init_register_light(self, light: "Light") -> Subgroup:
        """Creates neurons for the light source, if they don't already exist"""
        if self.light_source_ng is not None:
            Irr0_prev = self.light_source_ng.Irr0
            n_prev = self.light_source_ng.N
            # need to remove the old light source from the network
            self._remove_brian_object(self.light_source_ng)
        else:
            Irr0_prev = []
            n_prev = 0

        # create new one
        self.light_source_ng = NeuronGroup(
            n_prev + light.n, "Irr0: watt/meter**2", name="light_source"
        )
        if n_prev > 0:
            self.light_source_ng[:n_prev].Irr0 = Irr0_prev
        self._add_brian_object(self.light_source_ng)
        self.subgroup_idx_for_light[light] = slice(n_prev, n_prev + light.n)

        # remove and replace light_prop_syns for previous connections
        for light_prop_syn in self.light_prop_syns.values():
            self._remove_brian_object(light_prop_syn)
        prev_cxns = self.connections.copy()
        self.connections.clear()
        self.light_prop_syns.clear()
        for light, ldd, ng in prev_cxns:
            self.connect_light_to_ldd_for_ng(light, ldd, ng)
        assert prev_cxns == self.connections

    def register_light(self, light: "Light", ng: NeuronGroup):
        """Connects light to light-dependent devices already injected into this neuron group"""
        # create new connections for this light
        if ng not in self.lights_for_ng:
            self.lights_for_ng[ng] = set()
        self.lights_for_ng[ng].add(light)
        prev_injct_ldds = self.ldds_for_ng.get(ng, set())
        for ldd in prev_injct_ldds:
            self.connect_light_to_ldd_for_ng(light, ldd, ng)

    def source_for_light(self, light: "Light") -> Subgroup:
        """Returns the subgroup representing the given light source"""
        i = self.subgroup_idx_for_light[light]
        return self.light_source_ng[i]


registries: dict["CLSimulator", DeviceInteractionRegistry] = {}
"""Maps simulator to its registry"""


def registry_for_sim(sim: "CLSimulator") -> DeviceInteractionRegistry:
    """Returns the registry for the given simulator"""
    assert sim is not None
    if sim not in registries:
        registries[sim] = DeviceInteractionRegistry(sim)
    return registries[sim]
